/**CHeaderFile*************************************************************
  PackageName [bdd]
  Synopsis    [Package 'bdd' provides data structures and algorithms
               for the representation and manipulation of Boolean
               functions with ROBDDs. A hash table is used for quick
               search of nodes. Complement edges decreases the number
               of nodes. Function ITE is given for manipulation of
               boolean functions. An automatic garbage collection with
               a formulae counter is implemented.]

  FileName    [bddInt.h]
  Revision    [$Revision: 2 $]
  Date        [$Date: 2006-10-05 15:37:20 +0200 (Thu, 05 Oct 2006) $]
  Authors     [Robert Meolic (meolic@uni-mb.si),
               Ales Casar (casar@uni-mb.si)]
  Description [File bddInt.h contains declaration of internal
               data structures.]
  SeeAlso     [bdd.h]

  Copyright   [see bdd.h]
  ************************************************************************/

#ifndef _BDDINT
#define _BDDINT

#ifndef TCLOUT
#  define Mi_Out printf
   char outthread;
   int outflag;
   char *out;
#else
#  include <mi.h>
#  define printf Mi_Printf
#endif

#include "bdd.h"

/*-----------------------------------------------------------------------*/
/* Constant declarations                                                 */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Structure declarations                                                */
/*-----------------------------------------------------------------------*/

/* NODE TABLE = a hash table with chaining */
typedef struct {
  Bdd_Node **table;
  int size;            /* size of node table */
  int num;             /* number of all nodes in node table */
  int nf;              /* number of nonfortified nodes */
  int max;             /* maximal nodes in node table */
  int foa;             /* number of calls to Bdd_FoaNode */
  int add;             /* number of nodes generated */
  int compare;         /* number of comparisons made by Bdd_FoaNode */
  int generated;       /* number of free nodes generated */
  int garbage;         /* number of garbage collections */
} BddNodeTable;

/* FORMULAE TREE = a binary tree */
typedef struct BddFormula {
  struct BddFormula *l, *r;
  Bdd_Edge p;
  Bdd_String name;
} BddFormula;

/* CACHE LIST */
typedef struct BddCache {
  struct BddCache *next;
  Bdd_UserFunction gc;
} BddCache;

/* ITE CACHE */
typedef struct {
  Bdd_Node *f, *g, *h;
  Bdd_Edge result;
  Bdd_Boolean hmark;
  Bdd_Boolean ok;     /* FALSE = not valid node! */
} BddIteCache;

typedef struct {
  BddIteCache *table;
  int size;
  int search;
  int find;
  int overwrite;
} BddIteCacheTable;

/* EAX Cache */
typedef struct {
  Bdd_Node *f;
  Bdd_Variable *v;
  Bdd_Edge result;
  Bdd_Boolean fmark;
  Bdd_Boolean ok;     /* FALSE = not valid node! */
} BddEAxCache;

typedef struct {
  BddEAxCache *table;
  int size;
  int search;
  int find;
  int overwrite;
} BddEAxCacheTable;

/* RelOp Cache */
typedef struct {
  int key;            /* 0 = not valid node! */
  Bdd_Node *f;
  Bdd_Node *g;
  Bdd_Node *r;
  Bdd_Boolean fmark;
  Bdd_Boolean gmark;
  Bdd_Boolean rmark;
} BddRelOpCache;

typedef struct {
  BddRelOpCache *table;
  int size;
  int search;
  int find;
  int write;
  int overwrite;
} BddRelOpCacheTable;

/* RelOpComplex Cache */
typedef struct {
  int key;            /* 0 = not valid node! */
  Bdd_Node *f;
  Bdd_Node *g;
  Bdd_Node *r;
  Bdd_String name;
  Bdd_Boolean fmark;
  Bdd_Boolean gmark;
  Bdd_Boolean rmark;
} BddRelOpComplexCache;

typedef struct {
  BddRelOpComplexCache *table;
  int size;
  int search;
  int find;
  int write;
  int overwrite;
} BddRelOpComplexCacheTable;

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

extern Bdd_Variable *bddSymbolTree;

extern BddNodeTable bddNodeTable;

extern BddFormula *bddFormulaeTree;

extern BddCache *bddCacheList;

extern Bdd_Node *bddNewNodes;

extern Bdd_Node *bddFirstNewNode;

extern Bdd_Node *bddFreshNodes;

extern Bdd_Node *bddFreeNodes;

extern BddIteCacheTable bddIteCache;

extern BddEAxCacheTable bddEAxCache;

extern BddRelOpCacheTable bddRelOpCache;

extern BddRelOpComplexCacheTable bddRelOpComplexCache;

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Function prototypes                                                   */
/*-----------------------------------------------------------------------*/

extern Bdd_Boolean BddFindIteCache(
               Bdd_Edge a, Bdd_Edge b, Bdd_Edge c, Bdd_Edge *r);

extern void BddAddIteCache(
               Bdd_Edge a, Bdd_Edge b, Bdd_Edge c, Bdd_Edge r);

extern void BddIteGarbage();

extern void BddEAxGarbage();

extern void BddRelOpGarbage();

extern void BddRelOpComplexGarbage();

extern void BddWriteTable(int i);

extern void BddFindNode(Bdd_Variable *v, Bdd_Node *f, Bdd_Node *t);

extern void BddCheckTable();

extern void BddWriteTree();

extern int BddTreeMaxLevel();

extern int BddTreeMinLevel();

extern float BddTreeAvgLevel();

extern int BddTreeNumber();

/**AutomaticEnd***********************************************************/

#endif  /* _BDDINT */
