/**CFile*******************************************************************
  PackageName [bdd]
  Synopsis    [Package 'bdd' provides data structures and algorithms
               for the representation and manipulation of Boolean
               functions with ROBDDs. A hash table is used for quick
               search of nodes. Complement edges decreases the number
               of nodes. Function ITE is given for manipulation of
               boolean functions. An automatic garbage collection with
               a formulae counter is implemented.]

  FileName    [bddStat.c]
  Revision    [$Revision: 2 $]
  Date        [$Date: 2006-10-05 15:37:20 +0200 (Thu, 05 Oct 2006) $]
  Authors     [Robert Meolic (meolic@uni-mb.si),
               Ales Casar (casar@uni-mb.si)]
  Description [The file bddStat.c contains statistical functions.]
  SeeAlso     [bdd.h, bddInt.h]

  Copyright   [see bdd.h]
  ************************************************************************/

#include "bddInt.h"

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static void NodeNumber(Bdd_Node *f, int *i);

static void MaxLevel(Bdd_Node *f, int *max, int i);

static void AvgLevel(Bdd_Node *f, float *sum, int *n, int i);

static void NodeRepair(Bdd_Node *f);

static void WriteTree(Bdd_Variable *v);

static void MaxLevelTree(Bdd_Variable *v, int *max, int i);

static void MinLevelTree(Bdd_Variable *v, int *min, int i);

static void AvgLevelTree(Bdd_Variable *v, float *sum, int *n, int i);

static void TreeNumber(Bdd_Variable *v, int *n);

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

void
Bdd_SystemStat()
{
  fprintf(stdout,"  bddNodeTable.size = %d\n",bddNodeTable.size);
  fprintf(stdout,"  bddNodeTable.num = %d\n",bddNodeTable.num);
  fprintf(stdout,"  bddNodeTable.nf = %d\n",bddNodeTable.nf);
  fprintf(stdout,"  bddNodeTable.max = %d\n",bddNodeTable.max);
  fprintf(stdout,"  bddNodeTable.foa = %d\n",bddNodeTable.foa);
  fprintf(stdout,"  bddNodeTable.add = %d\n",bddNodeTable.add);
  fprintf(stdout,"  bddNodeTable.compare = %d\n",bddNodeTable.compare);
  fprintf(stdout,"  bddNodeTable.generated = %d\n",bddNodeTable.generated);
  fprintf(stdout,"  bddNodeTable.garbage = %d\n",bddNodeTable.garbage);
  fprintf(stdout,"  ListUsed = %d\n",Bdd_ListUsed());
  fprintf(stdout,"  ListMaxLength = %d\n",Bdd_ListMaxLength());
  fprintf(stdout,"  ListAvgLength = %f\n",Bdd_ListAvgLength());
  fprintf(stdout,"  CacheIteOk = %d\n",Bdd_CacheIteOk());
  fprintf(stdout,"  bddIteCache.find = %d\n",bddIteCache.find);
  fprintf(stdout,"  bddIteCache.overwrite = %d\n",bddIteCache.overwrite);
  fprintf(stdout,"  CacheRelOpOk = %d\n",Bdd_CacheRelOpOk());
  fprintf(stdout,"  bddRelOpCache.find = %d\n",bddRelOpCache.find);
  fprintf(stdout,"  bddRelOpCache.overwrite = %d\n",bddRelOpCache.overwrite);
  fprintf(stdout,"  CacheRelOpComplexOk = %d\n",Bdd_CacheRelOpComplexOk());
  fprintf(stdout,"  bddRelOpComplexCache.find = %d\n",
                                             bddRelOpComplexCache.overwrite);
  fprintf(stdout,"  bddRelOpComplexCache.overwrite = %d\n",
                                             bddRelOpComplexCache.find);
}

void
Bdd_FunctionStat(Bdd_Edge f)
{
  fprintf(stdout,"   NodeNumber = %d\n",Bdd_NodeNumber(f.p));
  fprintf(stdout,"   NodeMaxLevel = %d\n",Bdd_NodeMaxLevel(f.p));
  fprintf(stdout,"   NodeAvgLevel = %f\n",Bdd_NodeAvgLevel(f.p));
}

int
Bdd_NodeNumber(Bdd_Node *f)
{
  int i;

  i = 0;
  NodeNumber(f,&i);
  NodeRepair(f);
  return i;
}

int
Bdd_NodeMaxLevel(Bdd_Node *f)
{
  int max;

  max = 0;
  MaxLevel(f,&max,0);
  NodeRepair(f);
  return max;
}

float
Bdd_NodeAvgLevel(Bdd_Node *f)
{
  float sum;
  int n;

  sum = 0;
  n = 0;
  AvgLevel(f,&sum,&n,0);
  NodeRepair(f);
  return (sum/n);
}

int
Bdd_ListUsed()
{
  int i,n;

  n = 0;
  for (i=0; i<bddNodeTable.size; i++) {
    if (bddNodeTable.table[i] != NULL) {
      n++;
    }
  }
  return n;
}

int
Bdd_ListMaxLength()
{
  int i,n,max;
  Bdd_Node *sup;

  max = 0;
  for (i=0; i<bddNodeTable.size; i++) {
    if (bddNodeTable.table[i] != NULL) {
      n = 0;
      sup = bddNodeTable.table[i];
      while (sup != NULL) {
        n++;
        sup = sup->next;
      }
      if (n > max) max = n;
    }
  }
  return max;
}

float
Bdd_ListAvgLength()
{
  float sum;
  int i,n;
  Bdd_Node *sup;

  sum = 0;
  n = 0;
  for (i=0; i<bddNodeTable.size; i++) {
    if (bddNodeTable.table[i] != NULL) {
      n++;
      sup = bddNodeTable.table[i];
      while (sup != NULL) {
        sum = sum + 1;
        sup = sup->next;
      }
    }
  }
  return (sum/n);
}

int
Bdd_CacheIteOk()
{
  int i,n;

  n = 0;
  for (i=0; i<bddIteCache.size; i++)
    if (bddIteCache.table[i].ok) n++;

  return n;
}

int
Bdd_CacheRelOpOk()
{
  int i,n;

  n = 0;
  for (i=0; i<bddRelOpCache.size; i++)
    if (bddRelOpCache.table[i].key) n++;

  return n;
}

int
Bdd_CacheRelOpComplexOk()
{
  int i,n;

  n = 0;
  for (i=0; i<bddRelOpComplexCache.size; i++)
    if (bddRelOpComplexCache.table[i].key) n++;

  return n;
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

void
BddWriteTable(int i)
{
  Bdd_Node *sup;

  printf("[%d] ",i);
  sup = bddNodeTable.table[i];
  while (sup) {
    printf("*%x=(%s,%d,%x,%x,%x)",sup,sup->v->name,sup->v->order,sup->v,sup->f.p,sup->t.p);
    sup = sup->next;
    if (sup) printf(" ");
  }
  printf("\n");
}

/**Function****************************************************************
  Synopsis    [Function BddFindNode]
  Description [Used for testing. No mininization. No normalization.]
  SideEffects []
  SeeAlso     [Bdd_FoaNode]
  ************************************************************************/

void
BddFindNode(Bdd_Variable *v, Bdd_Node *f, Bdd_Node *t)
{
  int i;
  Bdd_Node *sup;

  printf("BddFindNode: ");
  for (i=0; i<bddNodeTable.size; i++) {

    sup = bddNodeTable.table[i];

    while (sup) {
      if (f == sup->f.p && t == sup->t.p && v == sup->v) {

        printf("[%d]",i);
      }
      sup = sup->next;
    }
  }
  printf("\n");
}

/**Function****************************************************************
  Synopsis    [Function BddCheckTable]
  Description [Used for testing.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
BddCheckTable()
{
  int i,j,num,max;
  Bdd_Node *sup,*sup1;

  max = bddNodeTable.size;

  printf("BddCheckTable:\n");

  for (i=0; i<max; i++) {

    sup = bddNodeTable.table[i];
    while (sup) {

      if (!(num = (int) (((unsigned int) sup->v->order + (unsigned int) sup->f.p +
		    (unsigned int) sup->t.p) % bddNodeTable.size))) num=1;

      if (i!=num) {
        printf("ERROR %x-(%s,%x,%x,%d,%x,%d) i=%d, num=%d\n",
                sup,sup->v->name,sup->v,sup->f.p,sup->f.mark,sup->t.p,sup->t.mark,i,num);
      }
      sup = sup->next;
    }
  }
  printf("OK\n");
}

/**Function****************************************************************
  Synopsis    [Function BddWriteTree]
  Description [Used for testing.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
BddWriteTree()
{
  WriteTree(bddSymbolTree);
  printf("\n");
}

/**Function****************************************************************
  Synopsis    [Function BddTreeMaxLevel]
  Description [Used for testing.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
BddTreeMaxLevel()
{
  int max;

  max = 0;
  MaxLevelTree(bddSymbolTree,&max,0);
  return max;
}

/**Function****************************************************************
  Synopsis    [Function BddTreeMinLevel]
  Description [Used for testing.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
BddTreeMinLevel()
{
  int min;

  min = BddTreeMaxLevel();
  MinLevelTree(bddSymbolTree,&min,0);
  return min;
}

/**Function****************************************************************
  Synopsis    [Function BddTreeAvgLevel]
  Description [Used for testing.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

float
BddTreeAvgLevel()
{
  float sum;
  int n;

  sum = 0;
  n = 0;
  AvgLevelTree(bddSymbolTree,&sum,&n,0);
  return (sum/n);
}

/**Function****************************************************************
  Synopsis    [Function BddTreeNumber]
  Description [Used for testing.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
BddTreeNumber()
{
  int n;

  n = 0;
  TreeNumber(bddSymbolTree,&n);
  return n;
}

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

static void
NodeNumber(Bdd_Node *f, int *i)
{

  if (f != NULL) {
    f->count = ~(f->count);
    (*i)++;
    if (f->f.p && (f->f.p->count >= 0) )
      NodeNumber(f->f.p,i);
    if (f->t.p && (f->t.p->count >= 0) )
      NodeNumber(f->t.p,i);
  }
}

static void
MaxLevel(Bdd_Node *f, int *max, int i)
{

  if (f != NULL) {
    f->count = ~(f->count);
    i++;
    if (i > *max) *max = i;
    if (f->f.p && (f->f.p->count >= 0) )
      MaxLevel(f->f.p,max,i);
    if (f->t.p && (f->t.p->count >= 0) )
      MaxLevel(f->t.p,max,i);
  }
}

static void
AvgLevel(Bdd_Node *f, float *sum, int *n, int i)
{

  if (f != NULL) {
    f->count = ~(f->count);
    i++;
    (*n)++;
    (*sum) = (*sum) + i;
    if (f->f.p && (f->f.p->count >= 0) )
      AvgLevel(f->f.p,sum,n,i);
    if (f->t.p && (f->t.p->count >= 0) )
      AvgLevel(f->t.p,sum,n,i);
  }
}

static void
NodeRepair(Bdd_Node *f)
{

  if (f != NULL) {
    f->count = ~(f->count);
    if (f->f.p && (f->f.p->count < 0) )
      NodeRepair(f->f.p);
    if (f->t.p && (f->t.p->count < 0) )
      NodeRepair(f->t.p);
  }
}

static void
WriteTree(Bdd_Variable *v)
{

  if (v != NULL) {
    printf("%s",v->name);
    printf(" (");
    WriteTree(v->l);
    printf(") (");
    WriteTree(v->r);
    printf(")");
  }
}

static void
MaxLevelTree(Bdd_Variable *v, int *max, int i)
{

  if (v == NULL) {
    if (i>*max) *max = i;
  } else {
    MaxLevelTree(v->l,max,i+1);
    MaxLevelTree(v->r,max,i+1);
  }
}

static void
MinLevelTree(Bdd_Variable *v, int *min, int i)
{

  if (v == NULL) {
    if (i<*min) *min = i;
  } else {
    MinLevelTree(v->l,min,i+1);
    MinLevelTree(v->r,min,i+1);
  }
}

static void
AvgLevelTree(Bdd_Variable *v, float *sum, int *n, int i)
{

  if (v != NULL) {
    i++;
    (*n)++;
    (*sum) = (*sum) + i;
    AvgLevelTree(v->l,sum,n,i);
    AvgLevelTree(v->r,sum,n,i);
  }
}

static void
TreeNumber(Bdd_Variable *v, int *n)
{

  if (v != NULL) {
    (*n)++;
    TreeNumber(v->l,n);
    TreeNumber(v->r,n);
  }
}
