/**CFile*******************************************************************
  PackageName [mc]
  Synopsis    [Package 'mc' provides a symbolic ACTLW model checker]

  FileName    [mcMain.c]
  Revision    [$Revision: 2 $]
  Date        [$Date: 2006-10-05 15:30:29 +0200 (Thu, 05 Oct 2006) $]
  Author      [Robert Meolic (meolic@uni-mb.si)]
  Description [File mcMain.c contains main functions of a symbolic
               ACTLW model checker.]
  SeeAlso     [mc.h, mcInt.h]

  Copyright   [see mc.h]
  ************************************************************************/

#include "mcInt.h"

/*-----------------------------------------------------------------------*/
/* Type declarations                                                     */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

int mc_status = 0;                  /* initialization status */
int mc_formulae;                    /* number of formulae */
Mc_Formula *mc_formulaTable;        /* table of formulae */

Bdd_Boolean trace = FALSE;          /* trace of ACTL parsing */
Bdd_Boolean xtrace = FALSE;         /* extended trace of ACTL parsing */
Bdd_Boolean explain = FALSE;        /* explain generated diagnostic */
Bdd_Boolean tracepath = FALSE;      /* full paths in diagnostic */
Bdd_Boolean diagnostic = FALSE;     /* generate linear w/c */
Bdd_Boolean automaton = FALSE;      /* generate w/c automaton */
Bdd_Boolean minimize = FALSE;       /* minimize w/c automaton */
Bdd_Boolean tree = FALSE;           /* generate tree-like w/c */

const int MC_SIMPLE = 0;
const int MC_TRACE = 1;
const int MC_XTRACE = 2;
const int MC_EXPLAIN = 4;
const int MC_TRACEPATH = 8;
const int MC_DIAGNOSTIC = 16;
const int MC_AUTOMATON = 32;
const int MC_MINIMIZE = 64;
const int MC_TREE = 128;

Bdd_Edge McTau;
McDiagnosticType McDgn;

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

#ifndef TCLOUT
   void Mc_Init(void) {}
#endif

/**Function****************************************************************
  Synopsis    [Function Mc_InitPkg.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Mc_InitPkg()
{
  if (mc_status == 1) {
    printf("\nModel checking package is already initialized. ");
    return;
  } else {
    mc_status = 1;
  }

  mc_formulae = 0;

  return;
}

/**Function****************************************************************
  Synopsis    [Function Mc_ExitPkg.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Mc_ExitPkg()
{
}

/**Function****************************************************************
  Synopsis    [Function Mc_AboutPkg]
  Description [Mc_AboutPkg reports release of Model checking package.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Mc_AboutPkg()
{
  printf("EST ");
  printf(EDITION);
  printf(", version ");
  printf(VERSION);
  printf(", ");
  printf("Model checking package");
  printf("\n");
}


/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Mc_AddNewFormula(Bdd_String name, Bdd_String formula)
{
  Mc_Formula *f;
  int i;
  int formnum;
  Bdd_Boolean OK;

  i = 0;
  if (mc_formulae > 0) {
    OK = FALSE;
    while (!OK && i<mc_formulae) {
      if (!strcmp(name,mc_formulaTable[i].name)) {
        OK = TRUE;
      } else {
        i++;
      }
    }
  }

  if (mc_formulae == 0) {
    mc_formulaTable = (Mc_Formula *) malloc (sizeof(Mc_Formula));
    f = mc_formulaTable;
    mc_formulae = 1;
    formnum = 0;
  } else {

    if (OK) {
      formnum = i;
      f = &mc_formulaTable[i];
      free(f->name);
      free(f->formula);
      printf(" (formula overwritten) ");
    } else {
      formnum = mc_formulae;
      mc_formulae++;
      mc_formulaTable = (Mc_Formula *) realloc (
                         mc_formulaTable, mc_formulae * sizeof(Mc_Formula));
      f = &mc_formulaTable[mc_formulae-1];
    }
  }

  f->name = strdup(name);
  f->formula = strdup(formula);

  return formnum;
}


/**Function****************************************************************
  Synopsis    []
  Description [FOA formula. If new formula is created its name will be
               composed of given prefix and a number.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Mc_FOANextFormula(Bdd_String f, Bdd_String prefix)
{
  int i;
  Bdd_String x;

  if (!prefix) {
    printf("\nError: Mc_FOANextFormula got null string!\n");
    return -1;
  }

  i = 0;
  while (i<mc_formulae) {
    if (!strcmp(f,mc_formulaTable[i].formula))
      return i;
    else
      i++;
  }

  x = (Bdd_String) malloc(127);
  sprintf(x,"%s%d",prefix,mc_formulae+1);

  i = Mc_AddNewFormula(x,f);

  free(x);
  return i;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Mc_FindFormula(Bdd_String name)
{
  int i = 0;

  if (!name) {
    printf("\nError: Mc_FindFormula got null string!\n");
    return -1;
  }
  while (i<mc_formulae) {
    if (!strcmp(name,mc_formulaTable[i].name)) {
      return i;
    } else {
      i++;
    }
  }

  return -1;
}


/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Mc_WriteKnownFormulae()
{
  int i;
  Bdd_String pom;

  printf("The list of formulae:\n");
  for(i=0;i<mc_formulae;i++) {
    pom = (Bdd_String) malloc(255);
    sprintf(pom,"%4d. %s == %s\n",i+1,mc_formulaTable[i].name,mc_formulaTable[i].formula);
    printf(pom);
    free(pom);
  }
  printf("\n");
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Mc_WriteFormula(Bdd_String ime, char type)
{
  int f;

  f = Mc_FindFormula(ime);
  if (f != -1) {

    if (type > 0) {
      printf("%s == ",mc_formulaTable[f].name);
    }
    printf("%s\n",mc_formulaTable[f].formula);

  } else {
    printf("Formula does not exist!\n");
  }
  printf("\n");
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckEEU.]
  Description [Strict EEU. Formula EE[{chi} phi U {chi'} phi']]
  SideEffects [EEU is not valid in deadlocked state.]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckEEU(Bdd_Edge D, Bdd_Edge a1, Bdd_Edge s1, Bdd_Edge a2,
            Bdd_Edge s2)
{
  Bdd_Edge sup,sup1,sup2,last,Z;

  /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */
  if (Bdd_Cmp(D,bdd_termFalse)) return bdd_termFalse; /* CAREFULLY !!! */

  sup2 = Bdd_RelOp(a2,s2,"#AND R2S",TRUE);
  sup2 = Bdd_RelOp(D,sup2,"#AND Ex xA xS",TRUE);   /* Z = ... -a2-> s2 */
  Bdd_Fortify(sup2.p);

  /* IF a1 == FALSE */
  if (Bdd_Cmp(a1,bdd_termFalse)) return sup2;

  sup1 = Bdd_ITE(s1,a1,bdd_termFalse);             
  Bdd_Fortify(sup1.p);

  Z = sup2;
  last = bdd_termTrue;
  while (!Bdd_Cmp(Z,last)) {

    Bdd_IncCounter();
    Bdd_Fresh(Z.p);

    last = Z;
    Z = Bdd_RelOp(Z,sup1,"#AND R2S",TRUE);
    Z = Bdd_RelOp(Z,D,"#AND Ex xA xS",TRUE);        /* ... -a1-> (s1*Z) */
    Z = Bdd_ITE(Z,bdd_termTrue,sup2);                   /* Z = Z + sup2 */
  }

  /*
  printf("EEU calculation finished.\n");
  */

  Bdd_Fortify(Z.p);
  return Z;                               /* where EEU is valid */
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckAAU.]
  Description [Strict AAU. Formula AA[{chi} phi U {chi'} phi']]
  SideEffects [AAU is not valid in deadlocked state.]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckAAU(Bdd_Edge D, Bdd_Edge a1, Bdd_Edge s1, Bdd_Edge a2,
            Bdd_Edge s2, Bdd_Edge *AAWfun, Bdd_Edge *AAFfun)
{
  Bdd_Edge r;

  /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */ /* CAREFULLY !!! */
  if (Bdd_Cmp(D,bdd_termFalse)) {
    *AAWfun = bdd_termTrue;
    *AAFfun = bdd_termFalse;
    return bdd_termFalse;
  }

  *AAWfun = Mc_CheckAAW(D,a1,s1,a2,s2);          /* where AAW is valid */
  *AAFfun = Mc_CheckAAF(D,a2,s2);                /* where AAF is valid */

  r =  Bdd_ITE(*AAWfun,*AAFfun,bdd_termFalse);

  Bdd_Fortify(r.p);
  return r;                                      /* where AAU is valid */
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckEEW.]
  Description [Strict EEW. Formula EE[{chi} phi W {chi'} phi']]
  SideEffects [EEW is valid in deadlocked state.]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckEEW(Bdd_Edge D, Bdd_Edge a1, Bdd_Edge s1, Bdd_Edge a2,
            Bdd_Edge s2, Bdd_Edge *EEUfun, Bdd_Edge *EEGfun)
{
  Bdd_Edge r;

  /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */ /* CAREFULLY !!! */
  if (Bdd_Cmp(D,bdd_termFalse)) {
    *EEUfun = bdd_termFalse;
    *EEGfun = bdd_termTrue;
    return bdd_termTrue;
  }

  *EEUfun = Mc_CheckEEU(D,a1,s1,a2,s2);          /* where EEU is valid */
  *EEGfun = Mc_CheckEEG(D,a1,s1);                /* where EEG is valid */

  r =  Bdd_ITE(*EEUfun,bdd_termTrue,*EEGfun);

  Bdd_Fortify(r.p);
  return r;                                      /* where EEW is valid */
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckAAW.]
  Description [Strict AAW. Formula AA[{chi} phi W {chi'} phi']]
  SideEffects [AAW is valid in deadlocked state.]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckAAW(Bdd_Edge D, Bdd_Edge a1, Bdd_Edge s1, Bdd_Edge a2,
            Bdd_Edge s2)
{
  Bdd_Edge sall,D1not,D2not_ras,D2not_sar;
  Bdd_Edge sup,sup1,sup2,last,Z;

  /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */
  if (Bdd_Cmp(D,bdd_termFalse)) return bdd_termTrue; /* CAREFULLY !!! */

  sup1 = Bdd_RelOp(D,bdd_termFalse,"ExA ExS",TRUE);
  sup2 = Bdd_RelOp(D,bdd_termFalse,"ExA ExR S2R",TRUE);
  sall = Bdd_ITE(sup1,bdd_termTrue,sup2);
  Bdd_Fortify(sall.p);                 /* sall = the set of all states */

  sup = Bdd_ITE(a1,s1,bdd_termFalse);
  sup = Bdd_RelOp(sup,bdd_termFalse,"R2S",TRUE);
  D1not = Bdd_ITE(sup,bdd_termFalse,D);
  Bdd_Fortify(D1not.p);                    /* D1not = delta_not(a1,s1) */ 

  sup = Bdd_ITE(a2,s2,bdd_termFalse);
  sup = Bdd_RelOp(sup,bdd_termFalse,"R2S",TRUE);
  D2not_ras = Bdd_ITE(sup,bdd_termFalse,D);
  D2not_sar = Bdd_RelOp(D2not_ras,bdd_termFalse,"R2S S2R",TRUE);
  Bdd_Fortify(D2not_ras.p);                /* D2not = delta_not(a2,s2) */ 
  Bdd_Fortify(D2not_sar.p);                /* D2not = delta_not(a2,s2) */ 

  Z = Bdd_RelOp(D1not,D2not_ras,"#AND Ex xA xS",TRUE);
  Bdd_Fortify(Z.p);

  last = bdd_termTrue;
  while (!Bdd_Cmp(Z,last)) {
    Bdd_IncCounter();
    Bdd_Fresh(Z.p);

    last = Z;
    sup = Bdd_RelOp(Z,D2not_sar,"#AND Ex xA xR S2R",TRUE);
    Z = Bdd_ITE(sup,bdd_termTrue,Z);
  }

  /*
  printf("AAW calculation finished.\n");
  */

  /* Z is now set of states where AAW is not valid */
  /* Z must be set of states where AAW is valid */

  Z = Bdd_ITE(Z,bdd_termFalse,sall);
  Bdd_Fortify(Z.p);

  return Z;                                        /* where AAW is valid */
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckEEX.]
  Description [EEX {chi} phi]
  SideEffects [EEX is not valid in deadlocked state.]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckEEX(Bdd_Edge D, Bdd_Edge a, Bdd_Edge s)
{
  Bdd_Edge r;

  /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */
  if (Bdd_Cmp(D,bdd_termFalse)) return bdd_termFalse;   /* CAREFULLY !!! */

  r =  Mc_CheckEEU(D,bdd_termFalse,bdd_termFalse,a,s);

  return r;
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckAAX.]
  Description [AAX {chi} phi]
  SideEffects [AAX is valid in deadlocked state.]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckAAX(Bdd_Edge D, Bdd_Edge a, Bdd_Edge s)
{
  Bdd_Edge r;

  /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */
  if (Bdd_Cmp(D,bdd_termFalse)) return bdd_termTrue;   /* CAREFULLY !!! */

  r =  Mc_CheckAAW(D,bdd_termFalse,bdd_termFalse,a,s);

  return r;
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckEEF.]
  Description [Strict EEF. Formula EEF {chi} phi]
  SideEffects [EEF is not valid in deadlocked state.]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckEEF(Bdd_Edge D, Bdd_Edge a, Bdd_Edge s)
{
  Bdd_Edge r;

  /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */
  if (Bdd_Cmp(D,bdd_termFalse)) return bdd_termFalse;   /* CAREFULLY !!! */

  r = Mc_CheckEEU(D,bdd_termTrue,bdd_termTrue,a,s);

  return r;
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckEEG.]
  Description [Strict EEG. Formula EEG {chi} phi]
  SideEffects [EEG is valid in deadlocked state.]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckEEG(Bdd_Edge D, Bdd_Edge a, Bdd_Edge s)
{
  Bdd_Edge sup,sup1,sup2,sd,last,Z;

  /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */   /* CAREFULLY !!! */
  if (Bdd_Cmp(D,bdd_termFalse)) return bdd_termTrue;

  sup1 = Bdd_RelOp(D,bdd_termFalse,"ExA ExS",TRUE);
  sup2 = Bdd_RelOp(D,bdd_termFalse,"ExA ExR S2R",TRUE);
  sd = Bdd_ITE(sup1,bdd_termFalse,sup2);
  Bdd_Fortify(sd.p);                 /* the set of all deadlocked states */

  Z = bdd_termTrue;
  last = bdd_termFalse;
  while (!Bdd_Cmp(Z,last)) {
    Bdd_IncCounter();
    Bdd_Fresh(Z.p);

    last = Z;
    sup = Bdd_ITE(a,s,bdd_termFalse);
    sup = Bdd_RelOp(sup,Z,"#AND R2S",TRUE);
    sup = Bdd_RelOp(sup,D,"#AND Ex xA xS",TRUE);
    Z = Bdd_ITE(sup,bdd_termTrue,sd);           /* add deadlocked states */
  }

  /*
  printf("EEG calculation finished.\n");
  */

  Bdd_Fortify(Z.p);
  return Z;                                        /* where EEG is valid */
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckAAF.]
  Description [Strict AAF. Formula AAF {chi} phi]
  SideEffects [AAF is not valid in deadlocked state.]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckAAF(Bdd_Edge D, Bdd_Edge a, Bdd_Edge s)
{
  Bdd_Edge sall,Dnot_ras,Dnot_sar;
  Bdd_Edge sup,sup1,sup2,sd_r,sd_s,last,Z;

  /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */
  if (Bdd_Cmp(D,bdd_termFalse)) return bdd_termFalse;   /* CAREFULLY !!! */

  sup1 = Bdd_RelOp(D,bdd_termFalse,"ExA ExS",TRUE);
  sup2 = Bdd_RelOp(D,bdd_termFalse,"ExA ExR S2R",TRUE);
  sall = Bdd_ITE(sup1,bdd_termTrue,sup2);
  Bdd_Fortify(sall.p);                   /* sall = the set of all states */

  sd_r = Bdd_ITE(sup1,bdd_termFalse,sup2);
  sd_s = Bdd_RelOp(sd_r,bdd_termFalse,"R2S",TRUE);
  Bdd_Fortify(sd_r.p);               /* the set of all deadlocked states */
  Bdd_Fortify(sd_s.p);               /* the set of all deadlocked states */

  sup = Bdd_ITE(a,s,bdd_termFalse);
  sup = Bdd_RelOp(sup,bdd_termFalse,"R2S",TRUE);
  Dnot_ras = Bdd_ITE(sup,bdd_termFalse,D);
  Dnot_sar = Bdd_RelOp(Dnot_ras,bdd_termFalse,"R2S S2R",TRUE);
  Bdd_Fortify(Dnot_ras.p);                      /* Dnot = delta_not(a,s) */ 
  Bdd_Fortify(Dnot_sar.p);                      /* Dnot = delta_not(a,s) */ 

  Z = bdd_termTrue;
  last = bdd_termFalse;
  while (!Bdd_Cmp(Z,last)) {
    Bdd_IncCounter();
    Bdd_Fresh(Z.p);

    last = Z;

    /* THERE ARE TWO STEPS IN ONE ITERATION */
    /* THIS IS BETER, YOU DO NOT NEED TO CALL R2S */

    sup = Bdd_RelOp(Z,Dnot_sar,"#AND Ex xA xR",TRUE);
    Z = Bdd_ITE(sup,bdd_termTrue,sd_s);         /* add deadlocked states */

    sup = Bdd_RelOp(Z,Dnot_ras,"#AND Ex xA xS",TRUE);
    Z = Bdd_ITE(sup,bdd_termTrue,sd_r);         /* add deadlocked states */
  }

  /*
  printf("AAF calculation finished.\n");
  */

  /* Z is now set of states where AF is not valid */
  /* Z must be set of states where AF is valid */

  Z = Bdd_ITE(Z,bdd_termFalse,sall);
  Bdd_Fortify(Z.p);

  return Z;                                        /* where AAF is valid */
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckAAG.]
  Description [Strict AAG. Formula AAG {chi} phi]
  SideEffects [AAG is valid in deadlocked state.]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckAAG(Bdd_Edge D, Bdd_Edge a, Bdd_Edge s)
{
  Bdd_Edge r;

  /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */   /* CAREFULLY !!! */
  if (Bdd_Cmp(D,bdd_termFalse)) return bdd_termTrue;

  r = Mc_CheckAAW(D,a,s,bdd_termFalse,bdd_termFalse);

  return r;
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckEU.]
  Description [E[phi {chi} U {chi'} phi']]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckEU(Bdd_Edge D, Bdd_Edge s1, Bdd_Edge a1, Bdd_Edge a2,
            Bdd_Edge s2)
{
  Bdd_Edge r;

  if (Bdd_Cmp(a2,bdd_termNull)) {

    a1 = Bdd_ITE(a1,bdd_termTrue,McTau);  
    Bdd_Fortify(a1.p);
    r =  Mc_CheckEEU(D,a1,s1,a1,s2);
    r = Bdd_ITE(r,s1,bdd_termFalse);
    r = Bdd_ITE(r,bdd_termTrue,s2);  
    Bdd_Fortify(r.p);

  } else {

    a1 = Bdd_ITE(a1,bdd_termTrue,McTau);  
    a2 = Bdd_ITE(McTau,bdd_termFalse,a2);
    Bdd_Fortify(a1.p);
    Bdd_Fortify(a2.p);
    r =  Mc_CheckEEU(D,a1,s1,a2,s2);
    r = Bdd_ITE(r,s1,bdd_termFalse);

  }

  return r;
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckAU.]
  Description [A[phi {chi} U {chi'} phi']]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckAU(Bdd_Edge D, Bdd_Edge s1, Bdd_Edge a1, Bdd_Edge a2,
            Bdd_Edge s2)
{
  Bdd_Edge r;
  Bdd_Edge sup1,sup2;

  if (Bdd_Cmp(a2,bdd_termNull)) {

    a1 = Bdd_ITE(a1,bdd_termTrue,McTau);
    Bdd_Fortify(a1.p);
    r =  Mc_CheckAAU(D,a1,s1,a1,s2,&sup1,&sup2);
    r = Bdd_ITE(r,s1,bdd_termFalse);
    r = Bdd_ITE(r,bdd_termTrue,s2);  
    Bdd_Fortify(r.p);

  } else {

    a1 = Bdd_ITE(a1,bdd_termTrue,McTau);  
    a2 = Bdd_ITE(McTau,bdd_termFalse,a2);
    Bdd_Fortify(a1.p);
    Bdd_Fortify(a2.p);
    r =  Mc_CheckAAU(D,a1,s1,a2,s2,&sup1,&sup2);
    r = Bdd_ITE(r,s1,bdd_termFalse);

  }

  return r;
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckEX.]
  Description [EX {chi} phi]
  SideEffects [EX {tau} phi is not valid in deadlocked state.]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckEX(Bdd_Edge D, Bdd_Edge a, Bdd_Edge s)
{
  Bdd_Edge r;

  if (Bdd_Cmp(a,McTau)) {

    /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */
    /* NOP */

    r =  Mc_CheckEEU(D,bdd_termFalse,bdd_termFalse,a,s);  /* Mc_CheckEEX(D,a,s) */

  } else {

    /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */
    /* NOP */

    a = Bdd_ITE(McTau,bdd_termFalse,a);
    Bdd_Fortify(a.p);
    r =  Mc_CheckEEU(D,bdd_termFalse,bdd_termFalse,a,s);  /* Mc_CheckEEX(D,a,s) */

  }

  return r;
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckAX.]
  Description [AX {chi} phi]
  SideEffects [AX {tau} phi is not valid in deadlocked state.]
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckAX(Bdd_Edge D, Bdd_Edge a, Bdd_Edge s)
{
  Bdd_Edge r;
  Bdd_Edge sup1,sup2,sd;

  sup1 = Bdd_RelOp(D,bdd_termFalse,"ExA ExS",TRUE);
  sup2 = Bdd_RelOp(D,bdd_termFalse,"ExA ExR S2R",TRUE);
  sd = Bdd_ITE(sup1,bdd_termFalse,sup2);
  Bdd_Fortify(sd.p);                            /* the set of all deadlocked states */

  if (Bdd_Cmp(a,McTau)) {

    /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */
    if (Bdd_Cmp(D,bdd_termFalse)) return bdd_termFalse;            /* CAREFULLY !!! */

    r =  Mc_CheckAAW(D,bdd_termFalse,bdd_termFalse,a,s);      /* Mc_CheckAAX(D,a,s) */
    r = Bdd_ITE(sd,bdd_termFalse,r);                    /* remove deadlocked states */

  } else {

    /* IF THERE ARE NO TRANSITIONS IN THE PROCESS !! */
    if (Bdd_Cmp(D,bdd_termFalse)) return bdd_termFalse;            /* CAREFULLY !!! */

    a = Bdd_ITE(McTau,bdd_termFalse,a);
    Bdd_Fortify(a.p);
    r =  Mc_CheckAAW(D,bdd_termFalse,bdd_termFalse,a,s);      /* Mc_CheckAAX(D,a,s) */
    r = Bdd_ITE(sd,bdd_termFalse,r);                    /* remove deadlocked states */

  }

  return r;
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckHME.]
  Description [phi <chi> phi', <chi> phi',
               phi <EPSILON> phi', <EPSILON> phi']
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckHME(Bdd_Edge D, Bdd_Edge s1, Bdd_Edge a, Bdd_Edge s2)
{
  Bdd_Edge r;

  if (Bdd_Cmp(s1,bdd_termNull) && !Bdd_Cmp(a,bdd_termNull)) {

    /* CAREFULLY!! */
    /* WE PREFER TO USE STRONG HME, WHERE APPROPRIATE */
    r = Mc_CheckEX(D,a,s2);

    /* THIS IS WEAK VARIANT FROM DeNicola, Vaandrager, 1990 */
    /*
    r = Mc_CheckEU(D,bdd_termTrue,bdd_termFalse,a,s2);
    */

  } else {

    if (Bdd_Cmp(s1,bdd_termNull)) s1 = bdd_termTrue;
    r = Mc_CheckEU(D,s1,bdd_termFalse,a,s2);

  }

  return r;
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckHMA.]
  Description [[chi] phi, [EPSILON] phi]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckHMA(Bdd_Edge D, Bdd_Edge a, Bdd_Edge s)
{
  Bdd_Edge sup1,sup2,sall,snot,r;

  if (Bdd_Cmp(D,bdd_termFalse)) {
    sall = bdd_termTrue;
  } else {
    sup1 = Bdd_RelOp(D,bdd_termFalse,"ExA ExS",TRUE);
    sup2 = Bdd_RelOp(D,bdd_termFalse,"ExA ExR S2R",TRUE);
    sall = Bdd_ITE(sup1,bdd_termTrue,sup2);
    Bdd_Fortify(sall.p);                         /* the set of all states */
  }

  snot = Bdd_ITE(s,bdd_termFalse,sall);
  Bdd_Fortify(snot.p);                      /* the set of states not in s */

  r = Mc_CheckHME(D,bdd_termNull,a,snot);
  r = Bdd_ITE(r,bdd_termFalse,sall);
  Bdd_Fortify(r.p);

  return r;
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckEF.]
  Description [EF phi]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckEF(Bdd_Edge D, Bdd_Edge s)
{
  Bdd_Edge r;

  r = Mc_CheckEEU(D,bdd_termTrue,bdd_termTrue,bdd_termTrue,s);
                                        /* Mc_CheckEEF(D,bdd_termTrue,s) */
  r = Bdd_ITE(r,bdd_termTrue,s);  
  Bdd_Fortify(r.p);

  return r;
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckEG.]
  Description [EG phi]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckEG(Bdd_Edge D, Bdd_Edge s)
{
  Bdd_Edge r;

  r = Mc_CheckEEG(D,bdd_termTrue,s);
  r = Bdd_ITE(r,s,bdd_termFalse);

  return r;
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckAF.]
  Description [AF phi]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckAF(Bdd_Edge D, Bdd_Edge s)
{
  Bdd_Edge r;

  r = Mc_CheckAAF(D,bdd_termTrue,s);
  r = Bdd_ITE(r,bdd_termTrue,s);  
  Bdd_Fortify(r.p);

  return r;
}

/**Function****************************************************************
  Synopsis    [Function Mc_CheckAG.]
  Description [AG phi]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Mc_CheckAG(Bdd_Edge D, Bdd_Edge s)
{
  Bdd_Edge r;

  r = Mc_CheckAAW(D,bdd_termTrue,s,bdd_termFalse,bdd_termFalse);
                                        /* Mc_CheckAAG(D,s,bdd_termTrue) */
  r = Bdd_ITE(r,s,bdd_termFalse);
  return r;
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/
