/**CFile*******************************************************************
  PackageName [pa]
  Synopsis    [Package 'pa' is a framework for implementation of 
               process algebrae similar to Milner's CCS and Hoare's CSP.]

  FileName    [paKod.c]
  Revision    [$Revision: 2 $]
  Date        [$Date: 2006-10-05 15:30:29 +0200 (Thu, 05 Oct 2006) $]
  Authors     [Robert Meolic (meolic@uni-mb.si)]
  Description [File paKod.c implements encoding and decoding of LTSs.]
  SeeAlso     [pa.h, paInt.h]

  Copyright   [see pa.h]
  ************************************************************************/

#include <math.h>
#include "paInt.h"

/*-----------------------------------------------------------------------*/
/* Structure declarations                                                */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static Bdd_Edge EncodeMinterm(Bdd_Edge *tabVar, int n, int i, int min);

static int DecodeSortBddA(Pa_Sort *sort, Bdd_Edge F,
                          Bdd_Boolean mark, int minA, int maxA,
                          Bdd_String *act, Bdd_Boolean write);

static int DecodeProcessBddR(Pa_Process *proc, Bdd_Edge F,
                             Bdd_Boolean mark, int minR, int maxR,
                             Bdd_Boolean write);

static int DecodeProcessBddRAS(Pa_Process *proc, Bdd_Edge F,
                               Bdd_Boolean mark,
                               int minR, int maxR, int minA, int maxA,
                               int minS, int maxS, Bdd_Boolean write);

static int DecodeCompBddR(Pa_Composition *comp, Bdd_Edge F,
                          Bdd_Boolean mark, int *decodingTableR,
                          Bdd_Boolean write);

static int DecodeCompBddRAS(Pa_Composition *comp, Bdd_Edge F,
                            Bdd_Boolean mark, int minA, int maxA,
                            int *decodingTableR, int *decodingTableS,
                            Bdd_Boolean write);

static int DecodeProcProcPairBddRP(
                           Pa_Process *p1, Pa_Process *p2, Bdd_Edge F,
                           Bdd_Boolean mark, int minR, int maxR,
                           int minP, int maxP, Bdd_Boolean write);

static int DecodeCompProcPairBddRP(
                           Pa_Composition *comp, Pa_Process *p, Bdd_Edge F,
                           Bdd_Boolean mark, int *decodingTableR,
                           int minP, int maxP, Bdd_Boolean write);

static int DecodeCompCompPairBddRP(
                           Pa_Composition *c1, Pa_Composition *c2, Bdd_Edge F,
                           Bdd_Boolean mark, int *decodingTableR,
                           int *decodingTableS, Bdd_Boolean write);

static int DecodeSortAction(Pa_Sort *s, int minA, int maxA, Bdd_String *act,
                            Bdd_Boolean write);

static int DecodeProcessState(Pa_Process *p, int minR, int maxR,
                              Bdd_Boolean write);

static int DecodeProcessTR(Pa_Process *p, int minR, int maxR,
                           int minA, int maxA, int minS, int maxS,
                           Bdd_Boolean write);

static int DecodeCompState(Pa_Composition *c, int *decodingTableR,
                           Bdd_Boolean write);

static int DecodeCompTR(Pa_Composition *c, int minA, int maxA,
                        int *decodingTableR, int *decodingTableS,
                        Bdd_Boolean write);

static int DecodeProcProcPair(Pa_Process *p1, Pa_Process *p2,
                              int minR, int maxR, int minP, int maxP,
                              Bdd_Boolean write);

static int DecodeCompProcPair(Pa_Composition *c, Pa_Process *p,
                              int *decodingTableR, int minP, int maxP,
                              Bdd_Boolean write);

static int DecodeCompCompPair(Pa_Composition *c1, Pa_Composition *c2,
                              int *decodingTableR, int *decodingTableS,
                              Bdd_Boolean write);

static int DecodeAction(Pa_Sort *s, int min, int max, int f, Bdd_String *act,
                        Bdd_Boolean write, int more);

static int DecodeState(Pa_Process *p, int min, int max, int f,
                       Bdd_Boolean write, Bdd_Boolean pname, int more);

static Bdd_Boolean ExtractProcTransition(Pa_Process *p, Bdd_Edge F,
                      Bdd_Boolean mark, int *minR, int *minA, int *minS);

static Bdd_Boolean ExtractCompTransition(Pa_Composition *comp, Bdd_Edge F,
                      Bdd_Boolean mark, int *minA, int *decodingTableR,
		      int *decodingTableS);

static Bdd_Boolean ExtractProcState(Pa_Process *proc, Bdd_Edge F,
                      Bdd_Boolean mark, int *minR);

static Bdd_Boolean ExtractCompState(Pa_Composition *comp, Bdd_Edge F,
                      Bdd_Boolean mark, int *decodingTableR);

static Bdd_Boolean ExtractSortAction(Pa_Sort *sort, Bdd_Edge F,
		      Bdd_Boolean mark, int *minA);

static void CompStates2ProcStates(Pa_Composition *c, int i,
                      int *decodingTableR, int min, int max, int f,
                      Bdd_Edge F, int procnum);

static void CompTR2ProcTR(Pa_Composition *c, int iR, int iS,
                      int *decodingTableR, int *decodingTableS, int minA,
                      int maxA, int minR, int maxR, int minS, int maxS,
                      int fA, int fR,int fS, Bdd_Edge FR, Bdd_Edge FS,
                      int procnum);

static void concat(Bdd_String *s1, const Bdd_String s2);

static void printstate(Bdd_String name);

/**AutomaticEnd***********************************************************/


/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description [V tabeli sortov poiscemo sort z danim imenom in ga kodiramo.
               Action 'TAU' has to be the first action in the table!]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_EncodeSort(Bdd_String name)
{
  Pa_Sort *s;
  Bdd_Edge var;
  Bdd_Edge *tabVar;
  Bdd_String x;
  int i,n,w;

  i = Pa_FindSort(name);
  if (i == -1) return;
  s = &pa_sortTable[i];

  /* dolocimo stevilo potrebnih spremenljivk za kodiranje akcij */
  n = (int)(1.9999999+log(s->numActions+1)/log(2.0));
  s->numVar = n;

  /* tvorimo spremenljivke in dolocimo utezi */
  tabVar = (Bdd_Edge *) malloc(n * sizeof(Bdd_Edge));
  x = (Bdd_String) malloc(127);
  w = 1;
  for (i=0; i<n; i++) {

    /* HERE IS GIVEN THE ORDER OF a,b VARIABLES */

    sprintf(x,"a%d",i);            /* format "a%d\0" */
    var = Bdd_FoaTerminal(x);
    var.p->v->weight = w;
    tabVar[i] = var;

    sprintf(x,"b%d",i);            /* format "b%d\0" */
    var = Bdd_FoaTerminal(x);
    var.p->v->weight = w;

    w = w * 2;
  }

  free(x);

  /* spremenljivka, ki doloca tip akcije */

  s->a0 = tabVar[0];

  /* MAKE BDD FOR ACTION 'TAU' */

  s->table[0].p = EncodeMinterm(tabVar, 0, n-1, 0);
  Bdd_Fortify(s->table[0].p.p);

  /* MAKE BDDs FOR OTHER ACTIONS */

  for (i=1; i<s->numActions; i++) {
    s->table[i].p = EncodeMinterm(tabVar, 2*i, n-1, 1);
    Bdd_Fortify(s->table[i].p.p);
  }

  free(tabVar);

  s->encoded = 1;
  PaSaveSortActions(s);
  return;
} 

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_EncodeProcess(Bdd_String name)
{
  Pa_Process *p;
  Bdd_Edge tau;
  Bdd_Edge var;
  Bdd_Edge d,e;
  Bdd_Edge states;
  Bdd_Edge *tabVarR;
  Bdd_Edge *tabVarS;
  Bdd_String x;
  int i,n,w;

  i = Pa_FindProcess(name);
  if (i == -1) {
    printf("Process does not exists.\n");
    return;
  }

  p = &pa_processTable[i];

  if (p->sort == -1) {
    printf("ERROR: Process has no sort!\n");
    return;
  }

  /* razsirimo prehode do drugih procesov */
  Pa_ExpandGates(p);

  /* po potrebi kodiramo sort */
  if (!pa_sortTable[p->sort].encoded) {
    /*
    printf("Encoding sort:\n");
    printf("  %s...",pa_sortTable[p->sort].name);
    */
    Pa_EncodeSort(pa_sortTable[p->sort].name);
    /*
    printf(" OK\n");
    */
  }

  /* poiscemo dogodek TAU - za kontrolo */
  tau = pa_sortTable[p->sort].table[0].p;

  /* dolocimo stevilo potrebnih spremenljivk za kodiranje stanj */
  if (p->numStates == 1) n = 1;
    else n = (int)(0.9999999+log(p->numStates)/log(2.0));
  p->numVar = n;

  /* tvorimo spremenljivke in dolocimo utezi */
  tabVarR = (Bdd_Edge *) malloc(n * sizeof(Bdd_Edge));
  tabVarS = (Bdd_Edge* ) malloc(n * sizeof(Bdd_Edge));
  x = (Bdd_String) malloc(127);
  w = 1;
  for (i=0; i<n; i++) {

    /* HERE IS GIVEN THE ORDER OF r,s,p,q VARIABLES */

    /* variable 'r' */
    sprintf(x,"r%d<%s>",i,name);  /* format "r%d%s\0" */
    var = Bdd_FoaTerminal(x);
    var.p->v->weight = w;
    tabVarR[i] = var;

    /* variable 's' */
    sprintf(x,"s%d<%s>",i,name);  /* format "s%d%s\0" */
    var = Bdd_FoaTerminal(x);
    var.p->v->weight = w;
    tabVarS[i] = var;

    /* variable 'p' */
    sprintf(x,"p%d<%s>",i,name);  /* format "p%d%s\0" */
    var = Bdd_FoaTerminal(x);
    var.p->v->weight = w;

    /* variable 'q' */
    sprintf(x,"q%d<%s>",i,name);  /* format "q%d%s\0" */
    var = Bdd_FoaTerminal(x);
    var.p->v->weight = w;

    w = w * 2;
  }

  free(x);

  /* zgradimo BDD za vsako stanje */
  for (i=0; i<p->numStates; i++) {
    p->tableS[i].bddR = EncodeMinterm(tabVarR, i, n-1, 0);
    p->tableS[i].bddS = EncodeMinterm(tabVarS, i, n-1, 0);
    Bdd_Fortify(p->tableS[i].bddR.p);
    Bdd_Fortify(p->tableS[i].bddS.p);
  }
  states = bdd_termFalse;

  for (i=0; i<p->numStates; i++) {
    if (strcmp(p->tableS[i].name,"")) {
      states = Bdd_ITE(states,bdd_termTrue,p->tableS[i].bddR);
    }
  }
  p->stateBDD = states;
  Bdd_Fortify(p->stateBDD.p);

  /* zgradimo prehajalno relacijo */
  d = bdd_termFalse;
  for (i=0; i<p->numTransitions; i++) {

    if (p->tableT[i].type) {
      e = Bdd_ITE(pa_sortTable[p->sort].a0,
                  pa_sortTable[p->sort].table[p->tableT[i].action].p,
                  bdd_termFalse);
    } else {
      e = Bdd_ITE(pa_sortTable[p->sort].a0,
                  bdd_termFalse,
                  pa_sortTable[p->sort].table[p->tableT[i].action].p);
    }

    e = Bdd_ITE(e,p->tableS[p->tableT[i].source].bddR,bdd_termFalse);
    e = Bdd_ITE(e,p->tableS[p->tableT[i].dest].bddS,bdd_termFalse);
    d = Bdd_ITE(d,bdd_termTrue,e);

  }
  Bdd_Fortify(d.p);
  p->d =d;

  /* zgradimo stab BDD */
  d = bdd_termTrue;
  for(i=0; i<p->numVar; i++) {
    e = Bdd_ITE(tabVarR[i],tabVarS[i],Bdd_NOT(tabVarS[i]));
    d = Bdd_ITE(d,e,bdd_termFalse);
  }
  Bdd_Fortify(d.p);
  p->stab =d;

  free(tabVarR);
  free(tabVarS);

  p->encoded = 1;
  PaSaveProcStates(p);
  Bdd_SaveFormula("D_",name,p->d);
  if (p->initial != -1) {
    Bdd_SaveFormula("Initial_",p->name,p->tableS[p->initial].bddR);
  } else {
    Bdd_SaveFormula("Initial_",p->name,bdd_termFalse);
  }
  Bdd_SaveFormula("S_",p->name,p->stateBDD);

  /* POMOZNI IZPISI */

  /*
  Pa_DecodeProcessStates(p,p->stateBDD,FALSE);
  Pa_DecodeProcessTR(p,p->d,FALSE);
  Pa_DecodeProcessTR(p,Bdd_ITE(tau,bdd_termFalse,p->d),FALSE);
  printf("Stevilo BDD vozlisc = %d.\n",Bdd_NodeNumber(p->d.p));

  printf("\n");
  */
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_EncodeMultiProcesses(Bdd_String plist)
{
  Bdd_String list,name;
  int i;

  if (!plist) {
    printf("\nError: Pa_EncodeMultiProcesses got null string!\n");
    return;
  }

  /* first process */
  list = strdup(plist);
  name = strtok(list," ");

  if (!name) {
    printf("\nError: Pa_EncodeMultiProcesses got empty string!\n");
    return;
  }

  i = Pa_FindProcess(name);
  if (i == -1) {
    printf(" Process %s doesn't exist\n",name);
  } else {
    Pa_EncodeProcess(name);
  }

  /* other procesees */

  name = strtok(NULL," ");

  while (name) {

    i = Pa_FindProcess(name);
    if (i == -1) {
      printf(" Process %s doesn't exist\n",name);
    } else {
      Pa_EncodeProcess(name);
    }

    name = strtok(NULL," ");
  }

  free(list);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_ExpandGates(Pa_Process *p)
{
  Pa_Process *q;
  Pa_Sort *ps;
  Pa_Sort *qs;
  int i,j,num;
  Bdd_String gate,state;
  Bdd_Boolean repeat;
  int state1,action,type,state2;

  ps = &pa_sortTable[p->sort];
  do {
    repeat = FALSE;

    for (i=0; i<p->numTransitions; i++) {
      gate = p->tableS[p->tableT[i].dest].name;

      if (Pa_IsGate(gate)) {
        num = Pa_FindProcess(gate);
        if (num == -1) {

          /* PROCESS DOES NOT EXISTS */
          /* IT WILL BE RECOGNIZED AS AN EMPTY PROCESS */

          concat(&p->tableS[p->tableT[i].dest].name,"<");
          concat(&p->tableS[p->tableT[i].dest].name,p->name); /* must become non-gate */
          concat(&p->tableS[p->tableT[i].dest].name,">");

        } else {

          q = &pa_processTable[num];
          qs = &pa_sortTable[q->sort];

          state = q->tableS[q->initial].name;
          num = Pa_FindStateProcess(p,state);

	  /* DEBUGGING */
	  /*
          printf("state = %s\n",state);
          printf("num = %d\n",num);
	  */

          if (num != -1) {

	    /*
            printf("DESTINATION WILL BE CHANGED\n");
	    */

            p->tableT[i].dest = num;

          } else {

	    /*
            printf("EXTERNAL PROCESS WILL BE INCLUDED\n");
	    */

            repeat = TRUE;

            /* COPY ACTIONS */
            for (j=0; j<qs->numActions; j++) {
              Pa_FOASortAction(ps,qs->table[j].name);
	    }

            /* COPY STATES */
            for (j=0; j<q->numStates; j++) {
              state = q->tableS[j].name;

              if (strcmp(state,"")) {
                if (Pa_IsGate(state)) {
                  Pa_FOAGateProcess(p,state);
                } else {
                  Pa_FOAStateProcess(p,state);
	        }
	      }
            }

            for (j=0; j<q->numTransitions; j++) {
              state1 = Pa_FindStateProcess(p,q->tableS[q->tableT[j].source].name);
              action = Pa_FindSortAction(ps,qs->table[q->tableT[j].action].name);
              type = q->tableT[j].type;
              state2 = Pa_FindStateProcess(p,q->tableS[q->tableT[j].dest].name);
              Pa_FOATransition(p,state1,action,type,state2);
	    }

            p->tableT[i].dest = Pa_FindStateProcess(p,q->tableS[q->initial].name);
	  }

	  /* DEBUGGING */
	  /*
          printf("Process after this change\n");
          Pa_WriteProcess(p->name,3);
          printf("\n");
	  */

        }
      }
    }

  } while (repeat);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_String
Pa_DecodeSortActions(Pa_Sort *s, Bdd_Edge F, Bdd_Boolean write)
{
  int i,maxAction;
  int number;
  Bdd_String act;

  act = strdup("");
  if(s->encoded) {

    maxAction = 1;
    for (i=0; i<s->numVar; i++) maxAction= maxAction * 2;
    number = DecodeSortBddA(s, F, bdd_termTrue.mark, 0, maxAction-1, &act, write);

  } else {
    printf("Sort %s is not encoded.\n",s->name);
  }

  return act;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_DecodeProcess(Bdd_String name, Bdd_Boolean write)
{
  Pa_Process *p;
  int i,maxAction,maxState;
  int number;

  i = Pa_FindProcess(name);
  if (i == -1) return;

  p = &pa_processTable[i];

  if(p->encoded) {
    printf("\nPROCESS %s\n",p->name); 
    printf("Initial state: ");
    printstate(p->tableS[p->initial].name);
    printf("\n");
    if (write) {
      printf("Transitions:\n");
    } else {
      printf("Number of transitions = %d.\n",number);
    }

    maxAction = 1;
    for (i=0; i<pa_sortTable[p->sort].numVar; i++) maxAction = maxAction * 2;
    maxState = 1;
    for (i=0; i<p->numVar; i++) maxState= maxState * 2;

    number = DecodeProcessBddRAS(p, p->d, bdd_termTrue.mark, 0, maxState-1,
                                 0, maxAction-1, 0, maxState-1, write);


  } else {
    printf("Process %s is not encoded.\n",p->name);
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeProcessStates(Pa_Process *p, Bdd_Edge F, Bdd_Boolean write)
{
  int i,maxState;
  int number;

  if(p->encoded) {

    maxState = 1;
    for (i=0; i<p->numVar; i++) maxState= maxState * 2;
    number = DecodeProcessBddR(p,F,bdd_termTrue.mark,0,maxState-1,write);

  } else {
    printf("Process %s is not encoded.\n",p->name);
    number = -1;
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeProcessTR(Pa_Process *p, Bdd_Edge F, Bdd_Boolean write)
{
  int i,maxAction,maxState;
  int number;

  if(p->encoded) {
    maxAction = 1;
    for (i=0; i<pa_sortTable[p->sort].numVar; i++) maxAction = maxAction * 2;
    maxState = 1;
    for (i=0; i<p->numVar; i++) maxState= maxState * 2;

    number = DecodeProcessBddRAS(p, F, bdd_termTrue.mark, 0, maxState-1,
                                 0, maxAction-1, 0, maxState-1, write);

  } else {
    printf("Process %s is not encoded.\n",p->name);
    number = -1;
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Pa_DecodeComposition(Bdd_String name, Bdd_Boolean write)
{
  Pa_Composition *c;
  int *decodingTableR;
  int *decodingTableS;
  int i,j,maxAction,maxState;
  int number;

  i = Pa_FindComposition(name);
  if (i == -1) return;

  c = &pa_compositionTable[i];

  decodingTableR = (int *) calloc (2 * c->numProcesses, sizeof(int));
  decodingTableS = (int *) calloc (2 * c->numProcesses, sizeof(int));

  maxAction = 1;
  for (i=0; i<pa_sortTable[c->sort].numVar; i++) maxAction = maxAction * 2;

  for (i=0; i<c->numProcesses; i++) {
    maxState = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++)
      maxState = maxState * 2;
    decodingTableR[2*i] = 0;
    decodingTableR[2*i+1] = maxState-1;
    decodingTableS[2*i] = 0;
    decodingTableS[2*i+1] = maxState-1;
  }

  number = DecodeCompBddRAS(c,c->transitionBDD,bdd_termTrue.mark,0,
                            maxAction-1,decodingTableR,decodingTableS,write);

  free(decodingTableR);
  free(decodingTableS);

  if (!write) printf("Stevilo prehodov = %d.\n",number);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeCompStates(Pa_Composition *c, Bdd_Edge F, Bdd_Boolean write)
{
  int *decodingTableR;
  int i,j,maxState;
  int number;

  decodingTableR = (int *) calloc (2 * c->numProcesses, sizeof(int));

  for (i=0; i<c->numProcesses; i++) {
    maxState = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++)
      maxState = maxState * 2;
    decodingTableR[2*i] = 0;
    decodingTableR[2*i+1] = maxState-1;
  }

  number = DecodeCompBddR(c, F, bdd_termTrue.mark, decodingTableR, write);

  free(decodingTableR);

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeCompTR(Pa_Composition *c, Bdd_Edge F, Bdd_Boolean write)
{
  int *decodingTableR;
  int *decodingTableS;
  int i,j,maxAction,maxState;
  int number;

  decodingTableR = (int *) calloc (2 * c->numProcesses, sizeof(int));
  decodingTableS = (int *) calloc (2 * c->numProcesses, sizeof(int));

  maxAction = 1;
  for (i=0; i<pa_sortTable[c->sort].numVar; i++) maxAction = maxAction * 2;

  for (i=0; i<c->numProcesses; i++) {
    maxState = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++)
      maxState = maxState * 2;
    decodingTableR[2*i] = 0;
    decodingTableR[2*i+1] = maxState-1;
    decodingTableS[2*i] = 0;
    decodingTableS[2*i+1] = maxState-1;
  }

  number = DecodeCompBddRAS(c, F, bdd_termTrue.mark, 0, maxAction-1,
                            decodingTableR, decodingTableS, write);

  free(decodingTableR);
  free(decodingTableS);

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeProcProcPair(Pa_Process *p1, Pa_Process *p2, Bdd_Edge F,
                      Bdd_Boolean write)
{
  int i,maxState1,maxState2;
  int number;

  if ((p1->encoded) && (p2->encoded)) {
    maxState1 = 1;
    for (i=0; i<p1->numVar; i++) maxState1 = maxState1 * 2;
    maxState2 = 1;
    for (i=0; i<p2->numVar; i++) maxState2 = maxState2 * 2;

    number = DecodeProcProcPairBddRP(p1, p2, F, bdd_termTrue.mark,
                                     0, maxState1-1, 0, maxState2-1, write);

  } else {
    printf("Processes are not encoded.\n");
    number = -1;
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeCompProcPair(Pa_Composition *c, Pa_Process *p, Bdd_Edge F,
                      Bdd_Boolean write)
{
  int i,j,maxState1,maxState2;
  int *decodingTableR;
  int number;

  decodingTableR = (int *) calloc (2 * c->numProcesses, sizeof(int));

  if(p->encoded) {
    maxState1 = 1;
    for (i=0; i<p->numVar; i++) maxState1 = maxState1 * 2;

    for (i=0; i<c->numProcesses; i++) {
      maxState2 = 1;
      for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++)
        maxState2 = maxState2 * 2;
      decodingTableR[2*i] = 0;
      decodingTableR[2*i+1] = maxState2-1;
    }

    number = DecodeCompProcPairBddRP(c, p, F, bdd_termTrue.mark,
                                 decodingTableR, 0, maxState1-1, write);

  } else {
    printf("Process is not encoded.\n");
    number = -1;
  }

  free(decodingTableR);

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_DecodeCompCompPair(Pa_Composition *c1, Pa_Composition *c2, Bdd_Edge F,
                      Bdd_Boolean write)
{
  int i,j,maxState;
  int *decodingTableR,*decodingTableS;
  int number;

  decodingTableR = (int *) calloc (2 * c1->numProcesses, sizeof(int));
  decodingTableS = (int *) calloc (2 * c2->numProcesses, sizeof(int));

  for (i=0; i<c1->numProcesses; i++) {
    maxState = 1;
    for (j=0; j<pa_processTable[c1->tableP[i]].numVar; j++)
      maxState = maxState * 2;
    decodingTableR[2*i] = 0;
    decodingTableR[2*i+1] = maxState-1;
  }

  for (i=0; i<c2->numProcesses; i++) {
    maxState = 1;
    for (j=0; j<pa_processTable[c2->tableP[i]].numVar; j++)
      maxState = maxState * 2;
    decodingTableS[2*i] = 0;
    decodingTableS[2*i+1] = maxState-1;
  }

  number = DecodeCompCompPairBddRP(c1, c2, F, bdd_termTrue.mark,
                                   decodingTableR, decodingTableS, write);

  free(decodingTableR);
  free(decodingTableS);

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Pa_ExtractProcTransition(Pa_Process *p, Bdd_Edge f)
{
  Bdd_Boolean OK;
  Bdd_Edge r,a,s,ras;
  int minR,minA,minS;

  minR = 0;
  minA = 0;
  minS = 0;

  OK = ExtractProcTransition(p,f,bdd_termTrue.mark,&minR,&minA,&minS);

  if (!OK) {
    printf("ExtractProcTransition return FALSE!\n");
    sleep(5);
    exit(1);
  }

  /*
  printf("Extract state R:");
  printf(p->tableS[minR].name);
  */

  r = p->tableS[minR].bddR;

  /*
  printf("\nExtract action A:");
  printf(pa_sortTable[p->sort].table[(minA)/2].name);
  if (minA % 2) printf("!, "); else printf("?, ");
  */

  a = pa_sortTable[p->sort].table[(minA)/2].p;
  if (minA % 2) {
    a = Bdd_ITE(pa_sortTable[p->sort].a0,a,bdd_termFalse);
  } else {
    a = Bdd_ITE(pa_sortTable[p->sort].a0,bdd_termFalse,a);
  }

  /*
  printf("\nExtract state S:");
  printf(p->tableS[minS].name);
  printf("\n");
  */

  s = p->tableS[minS].bddS;

  ras = Bdd_ITE(r,s,bdd_termFalse);
  ras = Bdd_ITE(ras,a,bdd_termFalse);
 
  return ras;
}


/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Pa_ExtractCompTransition(Pa_Composition *c, Bdd_Edge f)
{
  Bdd_Boolean OK;
  Bdd_Edge a,r,r1,s,s1,ras;
  int i;
  int minA;
  int *decodingTableR;
  int *decodingTableS;

  minA = 0;

  decodingTableR = (int *) calloc (2 * c->numProcesses, sizeof(int));
  decodingTableS = (int *) calloc (2 * c->numProcesses, sizeof(int));

  for (i=0; i<c->numProcesses; i++) {
    decodingTableR[2*i] = 0;     /* minR */
    decodingTableR[2*i+1] = -1;  /* maxR is not used */
    decodingTableS[2*i] = 0;     /* minS */
    decodingTableS[2*i+1] = -1;  /* maxS is not used */
  }

  OK = ExtractCompTransition(c,f,bdd_termTrue.mark,&minA,decodingTableR,
                             decodingTableS);

  if (!OK) {
    printf("ExtractCompTransition return FALSE!\n");
    sleep(5);
    exit(1);
  }

  r = bdd_termTrue;
  for (i=0; i<c->numProcesses; i++) {

    /*
    fprintf(stderr,"\nExtract state R:");
    fprintf(stderr,pa_processTable[c->tableP[i]].tableS[decodingTableR[2*i]].name);
    */

    r1 = pa_processTable[c->tableP[i]].tableS[decodingTableR[2*i]].bddR;
    r = Bdd_ITE(r,r1,bdd_termFalse);  /* r = r * r1 */
  }

  /*
  fprintf(stderr,"\nExtract action A:");
  fprintf(stderr," (c->sort = %d, minA = %d) ",c->sort,minA);
  fprintf(stderr," (pa_sortTable[c->sort].numActions = %d) ",pa_sortTable[c->sort].numActions);
  fprintf(stderr,pa_sortTable[c->sort].table[(minA)/2].name);
  if (minA % 2) fprintf(stderr,"!, "); else fprintf(stderr,"?, ");
  */

  a = pa_sortTable[c->sort].table[(minA)/2].p;
  if (minA % 2) {
    a = Bdd_ITE(pa_sortTable[c->sort].a0,a,bdd_termFalse);
  } else {
    a = Bdd_ITE(pa_sortTable[c->sort].a0,bdd_termFalse,a);
  }

  s = bdd_termTrue;
  for (i=0; i<c->numProcesses; i++) {

    /*
    fprintf(stderr,"\nExtract state S:");
    fprintf(stderr,pa_processTable[c->tableP[i]].tableS[decodingTableS[2*i]].name);
    */

    s1 = pa_processTable[c->tableP[i]].tableS[decodingTableS[2*i]].bddS;
    s = Bdd_ITE(s,s1,bdd_termFalse);  /* s = s * s1 */
  }

  /*
  printf("\n");
  */

  free(decodingTableR);
  free(decodingTableS);

  ras = Bdd_ITE(r,s,bdd_termFalse);
  ras = Bdd_ITE(ras,a,bdd_termFalse);
 
  return ras;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Pa_ExtractProcState(Pa_Process *p, Bdd_Edge f)
{
  Bdd_Boolean OK;
  Bdd_Edge r;
  int minR;

  minR = 0;

  OK = ExtractProcState(p,f,bdd_termTrue.mark,&minR);

  if (!OK) {
    printf("Pa_ExtractProcState got FALSE from ExtractProcState!\n");
    exit(1);
  }

  /*
  printf("Extract state R:");
  printf(p->tableS[minR].name);
  printf("\n");
  */

  r = p->tableS[minR].bddR;

  return r;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_EnumerateProcState(Pa_Process *p, Bdd_Edge f)
{
  Bdd_Boolean OK;
  int minR;

  minR = 0;

  OK = ExtractProcState(p,f,bdd_termTrue.mark,&minR);

  if (!OK) {
    printf("Pa_EnumerateProcState got FALSE from ExtractProcState!\n");
    exit(1);
  }

  return minR;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Pa_ExtractCompState(Pa_Composition *c, Bdd_Edge f)
{
  Bdd_Boolean OK;
  Bdd_Edge r,r1;
  int i;
  int *decodingTableR;

  decodingTableR = (int *) calloc (2 * c->numProcesses, sizeof(int));

  for (i=0; i<c->numProcesses; i++) {
    decodingTableR[2*i] = 0;     /* minR */
    decodingTableR[2*i+1] = -1;  /* maxR is not used */
  }

  OK = ExtractCompState(c,f,bdd_termTrue.mark,decodingTableR);

  if (!OK) {
    printf("Pa_ExtractCompState got FALSE from ExtractCompState!\n");
    exit(1);
  }

  /*
  printf("Extract state R:");
  */

  r = bdd_termTrue;
  for (i=0; i<c->numProcesses; i++) {
    r1 = pa_processTable[c->tableP[i]].tableS[decodingTableR[2*i]].bddR;
    /*
    printf(pa_processTable[c->tableP[i]].tableS[decodingTableR[2*i]].name);
    */
    r = Bdd_ITE(r,r1,bdd_termFalse);  /* r = r * r1 */
  }

  /*
  printf("\n");
  */
 
  free(decodingTableR);

  return r;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Pa_ExtractSortAction(Pa_Sort *s, Bdd_Edge f)
{
  Bdd_Boolean OK;
  Bdd_Edge r;
  int minA;

  minA = 0;

  OK = ExtractSortAction(s,f,bdd_termTrue.mark,&minA);

  if (!OK) {
    printf("Pa_ExtractSortAction got FALSE from ExtractSortAction!\n");
    exit(1);
  }

  /*
  printf("Extract action A:");
  printf(s->table[(minA)/2].name);
  if (minA % 2) printf("!"); else printf("?");
  printf("\n");
  */

  r = s->table[(minA)/2].p;

  return r;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Pa_EnumerateSortAction(Pa_Sort *s, Bdd_Edge f)
{
  Bdd_Boolean OK;
  int minA;

  minA = 0;

  OK = ExtractSortAction(s,f,bdd_termTrue.mark,&minA);

  if (!OK) {
    printf("Pa_EnumerateSortAction got FALSE from ExtractSortAction!\n");
    exit(1);
  }

  return minA;
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description [States are encoded with variables r]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Composition2ProcessR(Pa_Composition *comp, Bdd_Edge F, Bdd_Boolean mark,
                     int *decodingTableR, int procnum)
{
  Bdd_Variable *s;
  int i,old;
  char ch;
  Bdd_String pname;
  int j,fS;

  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) {
      fS = 1;
      for (j=0; j<pa_processTable[comp->tableP[0]].numVar; j++) fS = fS * 2;
      CompStates2ProcStates(comp,0,decodingTableR,decodingTableR[0],
                            decodingTableR[1],fS,bdd_termTrue,procnum);
    }
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'r': {
                  i = strspn(&s->name[1],"0123456789");                   
                  pname = strdup(&s->name[i+2]);
                  pname[strlen(pname)-1] = 0;
                  i = Pa_FindProcessComp(comp,pname);
                  free(pname);

                  old = decodingTableR[2*i+1]; /* maxR */
                  decodingTableR[2*i+1] =
                    decodingTableR[2*i+1] - s->weight; /* maxR - s->weight */
                  Composition2ProcessR(
                    comp,F.p->f,mark,decodingTableR,procnum);
                  decodingTableR[2*i+1] = old;

                  old = decodingTableR[2*i]; /* minR */
                  decodingTableR[2*i] =
                    decodingTableR[2*i] + s->weight; /* minR + s->weight */
                  Composition2ProcessR(
                    comp,F.p->t,mark,decodingTableR,procnum);
                  decodingTableR[2*i] = old;

                  break;
                }
    }
  }
}

/**Function****************************************************************
  Synopsis    []
  Description [Transition relation is encoded with variables r,a,s.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Composition2ProcessRAS(Pa_Composition *comp, Bdd_Edge F, Bdd_Boolean mark,
                    int minA, int maxA, int *decodingTableR,
                    int *decodingTableS, int procnum)
{
  Bdd_Variable *s;
  int i,old;
  char ch;
  Bdd_String pname;
  int j,fS,fA;

  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) {
      fS = 1;
      for (j=0; j<pa_processTable[comp->tableP[0]].numVar; j++) fS = fS * 2;
      fA = 1;
      for (j=1; j<pa_sortTable[comp->sort].numVar; j++) fA = fA * 2;
      CompTR2ProcTR(comp,0,0,decodingTableR,decodingTableS,minA,
                       maxA,decodingTableR[0],decodingTableR[1],
                       decodingTableS[0],decodingTableS[1],fA,fS,fS,
                       bdd_termTrue,bdd_termTrue,procnum);
    }
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'r': {
                  i = strspn(&s->name[1],"0123456789");
                  pname = strdup(&s->name[i+2]);
                  pname[strlen(pname)-1] = 0;
                  i = Pa_FindProcessComp(comp,pname);
                  free(pname);

                  old = decodingTableR[2*i+1]; /* maxR */
                  decodingTableR[2*i+1] =
                    decodingTableR[2*i+1] - s->weight; /* maxR - s->weight */
                  Composition2ProcessRAS(comp,F.p->f,mark,minA,maxA,
                                       decodingTableR,decodingTableS,procnum);
                  decodingTableR[2*i+1] = old;

                  old = decodingTableR[2*i]; /* minR */
                  decodingTableR[2*i] =
                    decodingTableR[2*i] + s->weight; /* minR + s->weight */
                  Composition2ProcessRAS(comp,F.p->t,mark,minA,maxA,
                                       decodingTableR,decodingTableS,procnum);
                  decodingTableR[2*i] = old;

                  break;
                }
      case 's': {
                  i = strspn(&s->name[1],"0123456789");
                  pname = strdup(&s->name[i+2]);
                  pname[strlen(pname)-1] = 0;
                  i = Pa_FindProcessComp(comp,pname);
                  free(pname);

                  old = decodingTableS[2*i+1]; /* maxR */
                  decodingTableS[2*i+1] =
                    decodingTableS[2*i+1] - s->weight; /* maxS - s->weight */
                  Composition2ProcessRAS(comp,F.p->f,mark,minA,maxA,
                                       decodingTableR,decodingTableS,procnum);
                  decodingTableS[2*i+1] = old;

                  old = decodingTableS[2*i]; /* minS */
                  decodingTableS[2*i] =
                    decodingTableS[2*i] + s->weight; /* minS + s->weight */
                  Composition2ProcessRAS(comp,F.p->t,mark,minA,maxA,
                                       decodingTableR,decodingTableS,procnum);
                  decodingTableS[2*i] = old;

                  break;
                }
      case 'a': {
                  Composition2ProcessRAS(comp,F.p->f,mark,minA,maxA-s->weight,
                                       decodingTableR,decodingTableS,procnum);

                  Composition2ProcessRAS(comp,F.p->t,mark,minA+s->weight,maxA,
                                       decodingTableR,decodingTableS,procnum);
                  break;
                }
    }
  }
}

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Edge
EncodeMinterm(Bdd_Edge *tabVar, int n, int i, int min)
{
  int w;

  if (i < min) return bdd_termTrue;
  w = tabVar[i].p->v->weight;
  if (w <= n)
    return Bdd_ITE(tabVar[i],EncodeMinterm(tabVar,n-w,i-1,min),bdd_termFalse);
  else
    return Bdd_ITE(tabVar[i],bdd_termFalse,EncodeMinterm(tabVar,n,i-1,min));
}

/**Function****************************************************************
  Synopsis    []
  Description [Actions are encoded with variables a]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeSortBddA(Pa_Sort *sort, Bdd_Edge F, Bdd_Boolean mark,
               int minA, int maxA, Bdd_String *act, Bdd_Boolean write)
{
  Bdd_Variable *s;
  char ch;
  int number;
  Bdd_String act1;

  number = 0;
  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) {
       act1 = strdup("");
       number = DecodeSortAction(sort,minA,maxA,&act1,write);
       concat(act,act1);
       free(act1);
    }
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'a':
        {
          number = number +
            DecodeSortBddA(sort,F.p->f,mark,minA,maxA-s->weight,act,write);
          number = number +
            DecodeSortBddA(sort,F.p->t,mark,minA+s->weight,maxA,act,write);
          break;
        }
    }
  }

  return number;
}


/**Function****************************************************************
  Synopsis    []
  Description [States are encoded with variables r]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeProcessBddR(Pa_Process *proc, Bdd_Edge F, Bdd_Boolean mark,
                  int minR, int maxR, Bdd_Boolean write)
{
  Bdd_Variable *s;
  char ch;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) number = DecodeProcessState(proc,minR,maxR,write);
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'r':
        {
          number = number +
            DecodeProcessBddR(proc,F.p->f,mark,minR,maxR-s->weight,write);
          number = number +
            DecodeProcessBddR(proc,F.p->t,mark,minR+s->weight,maxR,write);
          break;
        }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description [Transition relation is encoded with variables r,a,s]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeProcessBddRAS(Pa_Process *proc, Bdd_Edge F, Bdd_Boolean mark,
                    int minR, int maxR, int minA, int maxA,
                    int minS, int maxS, Bdd_Boolean write)
{
  Bdd_Variable *s;
  char ch;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark)
      number = DecodeProcessTR(proc,minR,maxR,minA,maxA,minS,maxS,write);
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'r': 
        {
          number = number + DecodeProcessBddRAS(proc,F.p->f,mark,
                                      minR,maxR-s->weight,
                                      minA,maxA,minS,maxS,write);
          number = number + DecodeProcessBddRAS(proc,F.p->t,mark,
                                      minR+s->weight,maxR,
                                      minA,maxA,minS,maxS,write);
          break;
        }

      case 'a':
        {
          number = number + DecodeProcessBddRAS(proc,F.p->f,mark,
                                      minR,maxR,
                                      minA,maxA-s->weight,
                                      minS,maxS,write);
          number = number + DecodeProcessBddRAS(proc,F.p->t,mark,
                                      minR,maxR,
                                      minA+s->weight,maxA,
                                      minS,maxS,write);
          break;
        }

      case 's': 
        {
          number = number + DecodeProcessBddRAS(proc,F.p->f,mark,
                                      minR,maxR,minA,maxA,
                                      minS,maxS-s->weight,write);
          number = number + DecodeProcessBddRAS(proc,F.p->t,mark,
                                      minR,maxR,minA,maxA,
                                      minS+s->weight,maxS,write);
          break;
        }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description [States are encoded with variables r]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompBddR(Pa_Composition *comp, Bdd_Edge F, Bdd_Boolean mark,
               int *decodingTableR, Bdd_Boolean write)
{
  Bdd_Variable *s;
  int i,old;
  char ch;
  Bdd_String pname;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) number = DecodeCompState(comp,decodingTableR,write);
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'r': {
                  i = strspn(&s->name[1],"0123456789");                   
                  pname = strdup(&s->name[i+2]);
                  pname[strlen(pname)-1] = 0;
                  i = Pa_FindProcessComp(comp,pname);
                  free(pname);

                  old = decodingTableR[2*i+1]; /* maxR */
                  decodingTableR[2*i+1] =
                    decodingTableR[2*i+1] - s->weight; /* maxR - s->weight */
                  number = number +
                    DecodeCompBddR(comp,F.p->f,mark,decodingTableR,write);
                  decodingTableR[2*i+1] = old;

                  old = decodingTableR[2*i]; /* minR */
                  decodingTableR[2*i] =
                    decodingTableR[2*i] + s->weight; /* minR + s->weight */
                  number = number +
                    DecodeCompBddR(comp,F.p->t,mark,decodingTableR,write);
                  decodingTableR[2*i] = old;

                  break;
                }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description [Transition relation is encoded with variables r,a,s.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompBddRAS(Pa_Composition *comp, Bdd_Edge F, Bdd_Boolean mark,
                    int minA, int maxA, int *decodingTableR,
                    int *decodingTableS, Bdd_Boolean write)
{
  Bdd_Variable *s;
  int i,old;
  char ch;
  Bdd_String pname;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) {
       number =
         DecodeCompTR(comp,minA,maxA,decodingTableR,decodingTableS,write);
    }
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'r': {
                  i = strspn(&s->name[1],"0123456789");
                  pname = strdup(&s->name[i+2]);
                  pname[strlen(pname)-1] = 0;
                  i = Pa_FindProcessComp(comp,pname);
                  free(pname);

                  old = decodingTableR[2*i+1]; /* maxR */
                  decodingTableR[2*i+1] =
                    decodingTableR[2*i+1] - s->weight; /* maxR - s->weight */
                  number = number +
                    DecodeCompBddRAS(comp,F.p->f,mark,minA,maxA,
                                     decodingTableR,decodingTableS,write);
                  decodingTableR[2*i+1] = old;

                  old = decodingTableR[2*i]; /* minR */
                  decodingTableR[2*i] =
                    decodingTableR[2*i] + s->weight; /* minR + s->weight */
                  number = number +
                    DecodeCompBddRAS(comp,F.p->t,mark,minA,maxA,
                                     decodingTableR,decodingTableS,write);
                  decodingTableR[2*i] = old;

                  break;
                }
      case 's': {
                  i = strspn(&s->name[1],"0123456789");
                  pname = strdup(&s->name[i+2]);
                  pname[strlen(pname)-1] = 0;
                  i = Pa_FindProcessComp(comp,pname);
                  free(pname);

                  old = decodingTableS[2*i+1]; /* maxR */
                  decodingTableS[2*i+1] =
                    decodingTableS[2*i+1] - s->weight; /* maxS - s->weight */
                  number = number +
                    DecodeCompBddRAS(comp,F.p->f,mark,minA,maxA,
                                     decodingTableR,decodingTableS,write);
                  decodingTableS[2*i+1] = old;

                  old = decodingTableS[2*i]; /* minS */
                  decodingTableS[2*i] =
                    decodingTableS[2*i] + s->weight; /* minS + s->weight */
                  number = number +
                    DecodeCompBddRAS(comp,F.p->t,mark,minA,maxA,
                                     decodingTableR,decodingTableS,write);
                  decodingTableS[2*i] = old;

                  break;
                }
      case 'a': {
                  number = number +
                    DecodeCompBddRAS(comp,F.p->f,mark,
                                     minA,maxA-s->weight,
                                     decodingTableR,decodingTableS,write);

                  number = number +
                    DecodeCompBddRAS(comp,F.p->t,mark,
                                     minA+s->weight,maxA,
                                     decodingTableR,decodingTableS,write);

                  break;
                }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeProcProcPairBddRP(Pa_Process *p1, Pa_Process *p2, Bdd_Edge F,
                        Bdd_Boolean mark, int minR, int maxR, int minP,
                        int maxP, Bdd_Boolean write)
{
  Bdd_Variable *s;
  char ch;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) {
      number = DecodeProcProcPair(p1,p2,minR,maxR,minP,maxP,write);
    }
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'r': {
                  number = number +
                    DecodeProcProcPairBddRP(p1,p2,F.p->f,mark,
                                            minR,maxR-s->weight,
                                            minP,maxP,write);
                  number = number +
                    DecodeProcProcPairBddRP(p1,p2,F.p->t,mark,
                                            minR+s->weight,maxR,
                                            minP,maxP,write);
                  break;
                }
      case 'p': {
                  number = number +
                    DecodeProcProcPairBddRP(p1,p2,F.p->f,mark,
                                            minR,maxR,
                                            minP,maxP-s->weight,write);
                  number = number +
                    DecodeProcProcPairBddRP(p1,p2,F.p->t,mark,
                                            minR,maxR,
                                            minP+s->weight,maxP,write);
                  break;
                }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompProcPairBddRP(Pa_Composition *comp, Pa_Process *p, Bdd_Edge F,
                        Bdd_Boolean mark, int *decodingTableR, int minP,
                        int maxP, Bdd_Boolean write)
{
  Bdd_Variable *s;
  Bdd_String pname;
  int i,old;
  char ch;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) {
      number = DecodeCompProcPair(comp,p,decodingTableR,minP,maxP,write);
    }
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'r': {
                  i = strspn(&s->name[1],"0123456789");
                  pname = strdup(&s->name[i+2]);
                  pname[strlen(pname)-1] = 0;
                  i = Pa_FindProcessComp(comp,pname);
                  free(pname);

                  old = decodingTableR[2*i+1]; /* maxR */
                  decodingTableR[2*i+1] =
                    decodingTableR[2*i+1] - s->weight; /* maxR - s->weight */
                  number = number +
                    DecodeCompProcPairBddRP(comp,p,F.p->f,mark,
                                            decodingTableR,minP,maxP,write);
                  decodingTableR[2*i+1] = old;

                  old = decodingTableR[2*i]; /* minR */
                  decodingTableR[2*i] =
                    decodingTableR[2*i] + s->weight; /* minR + s->weight */
                  number = number +
                    DecodeCompProcPairBddRP(comp,p,F.p->t,mark,
                                            decodingTableR,minP,maxP,write);
                  decodingTableR[2*i] = old;

                  break;
                }
      case 'p': {
                  number = number +
                    DecodeCompProcPairBddRP(comp,p,F.p->f,mark,
                                            decodingTableR,
                                            minP,maxP-s->weight,write);
                  number = number +
                    DecodeCompProcPairBddRP(comp,p,F.p->t,mark,
                                            decodingTableR,
                                            minP+s->weight,maxP,write);
                  break;
                }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompCompPairBddRP(Pa_Composition *c1, Pa_Composition *c2, Bdd_Edge F,
                        Bdd_Boolean mark, int *decodingTableR,
                        int *decodingTableS, Bdd_Boolean write)
{
  Bdd_Variable *s;
  Bdd_String pname;
  int i,old;
  char ch;
  int number;

  number = 0;
  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) {
      number = DecodeCompCompPair(c1,c2,decodingTableR,decodingTableS,write);
    }
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'r': {
                  i = strspn(&s->name[1],"0123456789");
                  pname = strdup(&s->name[i+2]);
                  pname[strlen(pname)-1] = 0;
                  i = Pa_FindProcessComp(c1,pname);
                  free(pname);

                  old = decodingTableR[2*i+1]; /* maxR */
                  decodingTableR[2*i+1] =
                    decodingTableR[2*i+1] - s->weight; /* maxR - s->weight */
                  number = number +
                    DecodeCompCompPairBddRP(c1,c2,F.p->f,mark,decodingTableR,
                                            decodingTableS,write);
                  decodingTableR[2*i+1] = old;

                  old = decodingTableR[2*i]; /* minR */
                  decodingTableR[2*i] =
                    decodingTableR[2*i] + s->weight; /* minR + s->weight */
                  number = number +
                    DecodeCompCompPairBddRP(c1,c2,F.p->t,mark,decodingTableR,
                                            decodingTableS,write);
                  decodingTableR[2*i] = old;

                  break;
                }

      case 'p': {
                  i = strspn(&s->name[1],"0123456789");
                  pname = strdup(&s->name[i+2]);
                  pname[strlen(pname)-1] = 0;
                  i = Pa_FindProcessComp(c2,pname);
                  free(pname);

                  old = decodingTableS[2*i+1]; /* maxP */
                  decodingTableS[2*i+1] =
                    decodingTableS[2*i+1] - s->weight; /* maxP - s->weight */
                  number = number +
                    DecodeCompCompPairBddRP(c1,c2,F.p->f,mark,decodingTableR,
                                            decodingTableS,write);
                  decodingTableS[2*i+1] = old;

                  old = decodingTableS[2*i]; /* minP */
                  decodingTableS[2*i] =
                    decodingTableS[2*i] + s->weight; /* minP + s->weight */
                  number = number +
                    DecodeCompCompPairBddRP(c1,c2,F.p->t,mark,decodingTableR,
                                            decodingTableS,write);
                  decodingTableS[2*i] = old;

                  break;
                }
    }
  }

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeSortAction(Pa_Sort *s, int minA, int maxA, Bdd_String *act, Bdd_Boolean write)
{
  int i,fA;
  int number;
  Bdd_String act1;

  fA = 1;
  for (i=1; i<s->numVar; i++) fA = fA * 2;

  act1 = strdup("");
  number = DecodeAction(s,minA,maxA,fA,&act1,write,0);

  /*
  concat(act,"(");
  concat(act,act1);
  concat(act,")");
  */

  *act = strdup(act1);
  free(act1);

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeProcessState(Pa_Process *p, int minR, int maxR, Bdd_Boolean write)
{
  int i,fR;
  int number;

  fR = 1;
  for (i=1; i<p->numVar; i++) fR = fR * 2;

  if (write) printf("(");
  number = DecodeState(p,minR,maxR,fR,write,FALSE,0);
  if (write) printf(")");

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeProcessTR(Pa_Process *p, int minR, int maxR,
                int minA, int maxA, int minS, int maxS, Bdd_Boolean write)
{
  int i,fA,fS;
  int nr,na,ns,number;
  Bdd_String act;

  fS = 1;
  fA = 1;

  for (i=1; i<p->numVar; i++) fS = fS * 2;
  for (i=1; i<pa_sortTable[p->sort].numVar; i++) fA = fA * 2;

  act = strdup("");

  if (write) printf("(");
  nr = DecodeState(p,minR,maxR,fS,write,FALSE,0);
  if (write) printf(") --- ");
  na = DecodeAction(&pa_sortTable[p->sort],minA,maxA,fA,&act,write,0);
  if (write) printf(" ---> (");
  ns = DecodeState(p,minS,maxS,fS,write,FALSE,0);
  if (write) printf(")\n");

  free(act);

  number = nr * na * ns;
  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompState(Pa_Composition *c, int *decodingTableR, Bdd_Boolean write)
{
  int i,j,fS;
  int number;

  number = 1;
  if (write) printf("(");

  for (i=0; i<c->numProcesses; i++) {
    fS = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++) fS = fS * 2;
    if (write) printf("(");
    number = number * DecodeState(&pa_processTable[c->tableP[i]],
             decodingTableR[2*i],decodingTableR[2*i+1],fS,write,TRUE,0);
    if (write) {
      printf(")");
      if (i<c->numProcesses-1) printf(",");
    }
  }

  if (write) printf(")");

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
CompStates2ProcStates(Pa_Composition *c, int i, int *decodingTableR,
                      int min, int max, int f, Bdd_Edge F, int procnum)
{
  int decoded = 0;
  int j;
  int fS;
  Pa_Process *proc;

  if (min == max) {
    if (min < pa_processTable[c->tableP[i]].numStates) {
      F = Bdd_ITE(F,pa_processTable[c->tableP[i]].tableS[min].bddR,
                  bdd_termFalse);
      if (i == c->numProcesses - 1) {
        decoded++;
      } else {
        i++;
        fS = 1;
        for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++) fS = fS * 2;
        CompStates2ProcStates(c,i,decodingTableR,decodingTableR[2*i],
                              decodingTableR[2*i+1],fS,F,procnum);
      }
    }
  } else {
    if ( (max - min) >= f ) {
      CompStates2ProcStates(c,i,decodingTableR,min+f,max,f/2,F,procnum);
      CompStates2ProcStates(c,i,decodingTableR,min,max-f,f/2,F,procnum);
    } else {
      CompStates2ProcStates(c,i,decodingTableR,min,max,f/2,F,procnum);
    }
  }

  if (decoded != 1) return;

  /*
  Pa_DecodeCompStates(c,F,TRUE);
  */

  proc = &pa_processTable[procnum];
  Pa_FOAComposedStateProcess(proc,F,proc->name);
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
CompTR2ProcTR(Pa_Composition *c, int iR, int iS, int *decodingTableR,
              int *decodingTableS, int minA, int maxA, int minR, int maxR,
              int minS, int maxS, int fA, int fR,int fS, Bdd_Edge FR,
              Bdd_Edge FS, int procnum)
{
  int decoded = 0;
  int j;
  Pa_Process *proc;
  int nr,ns;

  /*
  printf(
    "CompTR2ProcTR: minR=%d, maxR=%d, minA=%d, maxA=%d, minS=%d, maxS=%d\n",
    minR, maxR, minA, maxA, minS, maxS);
  */

  if (minR == maxR) {
    if (minR < pa_processTable[c->tableP[iR]].numStates) {
      FR = Bdd_ITE(FR,pa_processTable[c->tableP[iR]].tableS[minR].bddR,
                  bdd_termFalse);
      if (iR == c->numProcesses - 1) {
        decoded++;
      } else {
        iR++;
        fR = 1;
        for (j=0; j<pa_processTable[c->tableP[iR]].numVar; j++) fR = fR * 2;
        CompTR2ProcTR(c,iR,iS,decodingTableR,decodingTableS,
                      minA,maxA,decodingTableR[2*iR],decodingTableR[2*iR+1],
                      minS,maxS,fA,fR,fS,FR,FS,procnum);
      }
    }
  } else {
    if ( (maxR - minR) >= fR ) {
      CompTR2ProcTR(c,iR,iS,decodingTableR,decodingTableS,
                    minA,maxA,minR+fR,maxR,minS,maxS,fA,fR/2,fS,FR,FS,procnum);
      CompTR2ProcTR(c,iR,iS,decodingTableR,decodingTableS,
                    minA,maxA,minR,maxR-fR,minS,maxS,fA,fR/2,fS,FR,FS,procnum);
    } else {
      CompTR2ProcTR(c,iR,iS,decodingTableR,decodingTableS,
                    minA,maxA,minR,maxR,minS,maxS,fA,fR/2,fS,FR,FS,procnum);
    }
  }

  if (decoded != 1) return;

  if (minA == maxA) {
    if (minA <= 2 * pa_sortTable[c->sort].numActions) {
      decoded++;
    }
  } else {
    if ( (maxA - minA) >= fA ) {
      CompTR2ProcTR(c,iR,iS,decodingTableR,decodingTableS,
                    minA+fA,maxA,minR,maxR,minS,maxS,fA/2,fR,fS,FR,FS,procnum);
      CompTR2ProcTR(c,iR,iS,decodingTableR,decodingTableS,
                    minA,maxA-fA,minR,maxR,minS,maxS,fA/2,fR,fS,FR,FS,procnum);
    } else {
      CompTR2ProcTR(c,iR,iS,decodingTableR,decodingTableS,
                    minA,maxA,minR,maxR,minS,maxS,fA/2,fR,fS,FR,FS,procnum);
    }
  }

  if (decoded != 2) return;

  if (minS == maxS) {
    if (minS < pa_processTable[c->tableP[iS]].numStates) {
      FS = Bdd_ITE(FS,pa_processTable[c->tableP[iS]].tableS[minS].bddR,
                  bdd_termFalse);
      if (iS == c->numProcesses - 1) {
        decoded++;
      } else {
        iS++;
        fS = 1;
        for (j=0; j<pa_processTable[c->tableP[iS]].numVar; j++) fS = fS * 2;
        CompTR2ProcTR(c,iR,iS,decodingTableR,decodingTableS,
                      minA,maxA,minR,maxR,decodingTableS[2*iS],
                      decodingTableS[2*iS+1],fA,fR,fS,FR,FS,procnum);
        return;
      }
    }
  } else {
    if ( (maxS - minS) >= fS ) {
      CompTR2ProcTR(c,iR,iS,decodingTableR,decodingTableS,
                    minA,maxA,minR,maxR,minS+fS,maxS,fA,fR,fS/2,FR,FS,procnum);
      CompTR2ProcTR(c,iR,iS,decodingTableR,decodingTableS,
                    minA,maxA,minR,maxR,minS,maxS-fS,fA,fR,fS/2,FR,FS,procnum);
    } else {
      CompTR2ProcTR(c,iR,iS,decodingTableR,decodingTableS,
                    minA,maxA,minR,maxR,minS,maxS,fA,fR,fS/2,FR,FS,procnum);
    }
  }

  if (decoded != 3) return;

  /*
  Pa_DecodeCompTR(c,F,TRUE);
  */

  proc = &pa_processTable[procnum];
  nr = Pa_FOAComposedStateProcess(proc,FR,proc->name);
  ns = Pa_FOAComposedStateProcess(proc,FS,proc->name);

  /*
  fprintf(stderr,"Add transition: %d --%d=%d--> %d\n",nr,minA/2,minA%2,ns);
  */

  Pa_FOATransition(proc,nr,minA/2,minA%2,ns);
}


/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompTR(Pa_Composition *c, int minA, int maxA,
             int *decodingTableR, int *decodingTableS, Bdd_Boolean write)
{
  int i,j,fA,fS;
  int nr,na,ns,number;
  Bdd_String act;

  nr = 1;
  ns = 1;
  if (write) printf("(");

  for (i=0; i<c->numProcesses; i++) {
    fS = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++) fS = fS * 2;
    if (write) printf("(");
    nr = nr * DecodeState(&pa_processTable[c->tableP[i]],
         decodingTableR[2*i],decodingTableR[2*i+1],fS,write,TRUE,0);
    if (write) {
      printf(")");
      if (i<c->numProcesses-1) printf(",");
    }
  }

  if (write) printf(") --- ");

  fA = 1;
  act = strdup("");
  for (i=1; i<pa_sortTable[c->sort].numVar; i++) fA = fA * 2;
  na = DecodeAction(&pa_sortTable[c->sort],minA,maxA,fA,&act,write,0);

  if (write) printf(" ---> (");

  for (i=0; i<c->numProcesses; i++) {
    fS = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++) fS = fS * 2;
    if (write) printf("(");
    ns = ns * DecodeState(&pa_processTable[c->tableP[i]],
         decodingTableS[2*i],decodingTableS[2*i+1],fS,write,TRUE,0);
    if (write) {
      printf(")");
      if (i<c->numProcesses-1) printf(",");
    }
  }

  if (write) printf(")\n");

  free(act);

  number = nr * na * ns;
  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeProcProcPair(Pa_Process *p1, Pa_Process *p2, int minR, int maxR,
                   int minP, int maxP, Bdd_Boolean write)
{
  int i, fR, fP;
  int number;

  fR = 1;
  fP = 1;

  for (i=1; i<p1->numVar; i++) fR = fR * 2;
  for (i=1; i<p2->numVar; i++) fP = fP * 2;

  number = 1;
  if (write) printf("<(");
  number = number * DecodeState(p1,minR,maxR,fR,write,TRUE,0);
  if (write) printf("),(");
  number = number * DecodeState(p2,minP,maxP,fP,write,TRUE,0);
  if (write) printf(")>\n");

  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompProcPair(Pa_Composition *c, Pa_Process *p, int *decodingTableR,
                   int minP, int maxP, Bdd_Boolean write)
{
  int i,j,fS,fP;
  int nr,ns,number;

  nr = 1;
  if (write) printf("<<");

  for (i=0; i<c->numProcesses; i++) {
    fS = 1;
    for (j=0; j<pa_processTable[c->tableP[i]].numVar; j++) fS = fS * 2;
    if (write) printf("(");
    nr = nr * DecodeState(&pa_processTable[c->tableP[i]],
         decodingTableR[2*i],decodingTableR[2*i+1],fS,write,TRUE,0);
    if (write) printf("),");
  }

  if (write) printf(">,(");

  fP = 1;
  for (i=1; i<p->numVar; i++) fP = fP * 2;
  ns = DecodeState(p,minP,maxP,fP,write,TRUE,0);

  if (write) printf(")>\n");

  number = nr * ns;
  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeCompCompPair(Pa_Composition *c1, Pa_Composition *c2,
                   int *decodingTableR, int *decodingTableS,
                   Bdd_Boolean write)
{
  int i,j,fR,fS;
  int nr,ns,number;

  nr = 1;
  ns = 1;
  if (write) printf("<<");

  for (i=0; i<c1->numProcesses; i++) {
    fR = 1;
    for (j=0; j<pa_processTable[c1->tableP[i]].numVar; j++) fR = fR * 2;
    if (write) printf("(");
    nr = nr * DecodeState(&pa_processTable[c1->tableP[i]],
         decodingTableR[2*i],decodingTableR[2*i+1],fR,write,TRUE,0);
    if (write) printf("),");
  }

  if (write) printf(">,<");

  for (i=0; i<c2->numProcesses; i++) {
    fS = 1;
    for (j=0; j<pa_processTable[c2->tableP[i]].numVar; j++) fS = fS * 2;
    if (write) printf("(");
    ns = ns * DecodeState(&pa_processTable[c2->tableP[i]],
         decodingTableS[2*i],decodingTableS[2*i+1],fS,write,TRUE,0);
    if (write) printf("),");
  }

  if (write) printf(">>\n");

  number = nr * ns;
  return number;
}


/**Function****************************************************************
  Synopsis    []
  Description [Type variable (a0) should have weight = 1.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeAction(Pa_Sort *s, int min, int max, int f, Bdd_String *act,
             Bdd_Boolean write, int more)
{
  int number;
  Bdd_String act1;

  if (min > max) {
    fprintf(stdout,"DecodeAction ERROR!\n");
    fprintf(stdout,"name = %s, min = %d, max = %d\n",s->name,min,max);
    exit(1);
  }

  number = 0;
  if (min == max) {
    act1 = strdup("");
    if (min >= 2 * s->numActions) {
      if (more>0) concat(&act1,",");
      if (min % 2) concat(&act1,"!!"); else concat(&act1,"??");
    } else {
      if (min > 0) {
        if (more>0) concat(&act1,",");
        concat(&act1,s->table[(min)/2].name);
        if (min % 2) concat(&act1,"!"); else concat(&act1,"?");
      } else {
        if (more>0) concat(&act1,",");
        concat(&act1,"TAU");
      }

      number = 1;
    }

    if (write) printf(act1);
    concat(act,act1);
    free(act1);

  } else {

    if ( (max - min) >= f ) {
      number = number + DecodeAction(s,min+f,max,f/2,act,write,more);
      number = number + DecodeAction(s,min,max-f,f/2,act,write,more+1);
    } else {
     number = number + DecodeAction(s,min,max,f/2,act,write,more);
    }

  }
  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
DecodeState(Pa_Process *p, int min, int max, int f, Bdd_Boolean write,
            Bdd_Boolean pname, int more)
{
  int number;
  Bdd_String init;

  if (min > max) {
    printf("DecodeState ERROR!\n");
    printf("name = %s, min = %d, max = %d\n",p->name,min,max);
    exit(1);
  }

  number = 0;
  if (min == max) {

    if (min >= p->numStates) {
      if (write) printf("??");
    } else {
      if (write) {
        if (more>0) printf(")(");
        if (pname) {
          printf(p->tableS[min].name);
        } else {
          printstate(p->tableS[min].name);
	}
      }
      number = 1;
    }

  } else {

    if ( (max - min) >= f ) {
      number = number + DecodeState(p,min+f,max,f/2,write,pname,more);
      number = number + DecodeState(p,min,max-f,f/2,write,pname,more+1);
    } else {
      number = number + DecodeState(p,min,max,f/2,write,pname,more);
    }

  }
  return number;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Boolean
ExtractProcTransition(Pa_Process *p, Bdd_Edge F, Bdd_Boolean mark,
                      int *minR, int *minA, int *minS)
{
  Bdd_Variable *s;
  static char ch;    /* STATIC TO MINIMIZE STACK PROBLEMS */

  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) return TRUE; else return FALSE;
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'r':
        {
          if (ExtractProcTransition(p,F.p->f,mark,minR,minA,minS)) {
            return TRUE;
	  }

          *minR = *minR + s->weight;
          return ExtractProcTransition(p,F.p->t,mark,minR,minA,minS);

          break;
        }

      case 'a':
        {
          if (ExtractProcTransition(p,F.p->f,mark,minR,minA,minS)) {
            return TRUE;
	  }

          *minA = *minA + s->weight;
          return ExtractProcTransition(p,F.p->t,mark,minR,minA,minS);

          break;
        }

      case 's':
        {
          if (ExtractProcTransition(p,F.p->f,mark,minR,minA,minS)) {
            return TRUE;
	  }

          *minS = *minS + s->weight;
          return ExtractProcTransition(p,F.p->t,mark,minR,minA,minS);

          break;
        }

    }
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Boolean
ExtractCompTransition(Pa_Composition *comp, Bdd_Edge F, Bdd_Boolean mark,
                    int *minA, int *decodingTableR, int *decodingTableS)
{
  int i;
  Bdd_Variable *s;
  static char ch;           /* STATIC TO MINIMIZE STACK PROBLEMS */
  static Bdd_String pname;  /* STATIC TO MINIMIZE STACK PROBLEMS */

  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) return TRUE; else return FALSE;
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'r':
        {
          i = strspn(&s->name[1],"0123456789");
          pname = strdup(&s->name[i+2]);
          pname[strlen(pname)-1] = 0;
          i = Pa_FindProcessComp(comp,pname);
          free(pname);

          if (ExtractCompTransition(comp,F.p->f,mark,
                                    minA,decodingTableR,decodingTableS)) {
            return TRUE;
	  }

          decodingTableR[2*i] =
            decodingTableR[2*i] + s->weight; /* minR + s->weight */

          return ExtractCompTransition(comp,F.p->t,mark,minA,
                                       decodingTableR,decodingTableS);

          break;
        }

      case 'a':
        {
          if (ExtractCompTransition(comp,F.p->f,mark,
				    minA,decodingTableR,decodingTableS)) {
            return TRUE;
	  }


          *minA = *minA + s->weight;
          return ExtractCompTransition(comp,F.p->t,mark,
                                       minA,decodingTableR,decodingTableS);

          break;
        }

      case 's':
        {
          i = strspn(&s->name[1],"0123456789");
          pname = strdup(&s->name[i+2]);
          pname[strlen(pname)-1] = 0;
          i = Pa_FindProcessComp(comp,pname);
          free(pname);

          if (ExtractCompTransition(comp,F.p->f,mark,
                                    minA,decodingTableR,decodingTableS)) {
            return TRUE;
	  }

          decodingTableS[2*i] =
            decodingTableS[2*i] + s->weight; /* minS + s->weight */

          return ExtractCompTransition(comp,F.p->t,mark,minA,
                                       decodingTableR,decodingTableS);

          break;
        }
    }
  }
}

/**Function****************************************************************
  Synopsis    []
  Description [States are encoded with variables r]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Boolean
ExtractProcState(Pa_Process *proc, Bdd_Edge F, Bdd_Boolean mark, int *minR)
{
  Bdd_Variable *s;
  static char ch;   /* STATIC TO MINIMIZE STACK PROBLEMS */

  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) return TRUE; else return FALSE;
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'r':
        {
          if (ExtractProcState(proc,F.p->f,mark,minR)) {
            return TRUE;
	  }

          *minR = *minR + s->weight;
          return ExtractProcState(proc,F.p->t,mark,minR);

          break;
        }
    }
  }
}

/**Function****************************************************************
  Synopsis    []
  Description [States are encoded with variables r]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Boolean
ExtractCompState(Pa_Composition *comp, Bdd_Edge F, Bdd_Boolean mark,
                 int *decodingTableR)
{
  int i;
  Bdd_Variable *s;
  static char ch;           /* STATIC TO MINIMIZE STACK PROBLEMS */
  static Bdd_String pname;  /* STATIC TO MINIMIZE STACK PROBLEMS */

  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) return TRUE; else return FALSE;
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'r': {
                  i = strspn(&s->name[1],"0123456789");                   
                  pname = strdup(&s->name[i+2]);
                  pname[strlen(pname)-1] = 0;
                  i = Pa_FindProcessComp(comp,pname);
                  free(pname);

                  if (ExtractCompState(comp,F.p->f,mark,decodingTableR)) {
                    return TRUE;
	          }

                  decodingTableR[2*i] =
                    decodingTableR[2*i] + s->weight; /* minR + s->weight */

                  return ExtractCompState(comp,F.p->t,mark,decodingTableR);

                  break;
      }
    }
  }
}

/**Function****************************************************************
  Synopsis    []
  Description [Actions are encoded with variables a]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Boolean
ExtractSortAction(Pa_Sort *sort, Bdd_Edge F, Bdd_Boolean mark, int *minA)
{
  Bdd_Variable *s;
  static char ch;  /* STATIC TO MINIMIZE STACK PROBLEMS */

  if (F.mark) mark = (!mark);
  if (F.p == bdd_termTrue.p) {
    if (!mark) return TRUE; else return FALSE;
  } else {
    s = F.p->v;
    ch = s->name[0];
    switch (ch) {
      case 'a':
        {
          if (ExtractSortAction(sort,F.p->f,mark,minA)) {
            return TRUE;
	  }

          *minA = *minA + s->weight;
          return ExtractSortAction(sort,F.p->t,mark,minA);

          break;
        }
    }
  }
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
concat(Bdd_String *s1, const Bdd_String s2)
{
   *s1 = (char*) realloc(*s1,strlen(*s1)+strlen(s2)+1);
   strcat(*s1,s2);
}


/**Function****************************************************************
  Synopsis    [Function printstate]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
printstate(Bdd_String name) {
  int i;
  Bdd_String shortname;

  i = strcspn(name,"<");
  shortname = (char *) malloc(i+1);
  memcpy(shortname,name,i);
  shortname[i]=0;
  printf(shortname);
  free(shortname);
}
