/**CFile*******************************************************************
  PackageName [versis]
  Synopsis    [Package 'versis' provides basic tools for formal
               verification of concurrent systems]

  FileName    [versisComp.c]
  Revision    [$Revision: 2 $]
  Date        [$Date: 2006-10-05 15:30:29 +0200 (Thu, 05 Oct 2006) $]
  Authors     [Robert Meolic (meolic@uni-mb.si),
               Mirjam Sepesy Maucec (mirjam.sepesy@uni-mb.si),
               Tatjana Kapus (kapus@uni-mb.si)]
  Description [File versisComp.c provides parallel composition.]
  SeeAlso     [versis.h, versisInt.h]

  Copyright   [see versis.h]
  ************************************************************************/

#include "versisInt.h"

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static void Compose(Pa_Composition *c, int type);

static void Normalization(Pa_Composition *c);

static int wordNumber(Bdd_String s);

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description [0 - CCS compose general,
               1 - CCS compose given non-restricted actions,
               2 - CCS compose given restricted actions,
               3 - multi-way general,
               4 - multi-way given non-restricted actions,
               5 - multi-way given restricted actions,
               6 - fundamental-mode multi-way general,
               7 - fundamental-mode multi-way given non-restricted actions,
               8 - fundamental-mode multi-way given restricted actions]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Versis_Compose(Bdd_String name, int type, Bdd_String process,
               Bdd_String actions)
{
  int i,k;
  int numP,numA,numcomp;
  int token;
  Pa_Composition *c;
  Bdd_String sup,sup_next;
  Bdd_String s;
  Bdd_Edge trc,f1,f2,f3;
  clock_t ura1, ura2;

  ura1 = clock();

  numP = wordNumber(process);
  numA = wordNumber(actions);

  i = pa_compositions;

  numcomp = Pa_AddNewComposition(name,type,numP,numA);

  if (pa_compositions == i) {
    printf(" (composition overwritten) ");
  }
  c = &pa_compositionTable[numcomp];

  /* PROCESSES */

  /*
  fprintf(stderr,"Compose: %s\n",process);
  */

  sup = (Bdd_String) strdup(process);

  token = strcspn(sup," ");
  s = (Bdd_String) malloc(token+1);
  strncpy(s,sup,token);                 /* name of first process */
  s[token] = '\0';

  i = Pa_FindProcess(s);
  free(s);
 
  if (i == -1) {
    Pa_RemoveComposition(numcomp);
    printf(" Process %%%d does not exist.\n",1);
    return;
  }
  if (!pa_processTable[i].encoded) {
    Pa_RemoveComposition(numcomp);
    printf(" Process %%%d is not encoded.\n",1);
    return;
  }
  c->tableP[0] = i;                     /* number of first process */
  c->sort = pa_processTable[i].sort;

  k = 1;
  while (k < numP) {
    while (sup[token] == ' ') token = token + 1;
    sup_next = (Bdd_String) strdup(&sup[token]);
    free(sup);
    sup = sup_next;

    token = strcspn(sup," ");
    s = (Bdd_String) malloc(token+1);
    strncpy(s,sup,token);               /* name of next process */
    s[token] = '\0';

    i = Pa_FindProcess(s);
    free(s);

    if (i == -1) {
      Pa_RemoveComposition(numcomp);
      printf(" Process %%%d does not exist.\n",k*1);
      return;
    }
    if (!pa_processTable[i].encoded) {
      Pa_RemoveComposition(numcomp);
      printf(" Process %%%d is not encoded.\n",k+1);
      return;
    }
    c->tableP[k++] = i;
    if (c->sort != pa_processTable[i].sort) {
      Pa_RemoveComposition(numcomp);
      printf(" Processes have different sorts!\n");
      return;
    }
  }
  free(sup);

  /* ACTIONS */

  if ( actions ) {

    sup = (Bdd_String) strdup(actions);

    token = strcspn(sup," ");
    s = (Bdd_String) malloc(token+1);
    strncpy(s,sup,token);                 /* name of first action */
    s[token] = '\0';

    i = Pa_FindSortAction(&pa_sortTable[c->sort],s);
    free(s);
 
    if (i == -1) {
      Pa_RemoveComposition(numcomp);
      printf(" Action %%%d does not exist.\n",1);
      return;
    }

    c->tableA[0] = i;  /* number of first action */

    k = 1;
    while (k < numA) {
      while (sup[token] == ' ') token = token + 1;
      sup_next = (Bdd_String) strdup(&sup[token]);
      free(sup);
      sup = sup_next;

      token = strcspn(sup," ");
      s = (Bdd_String) malloc(token+1);
      strncpy(s,sup,token);                /* name of next action */
      s[token] = '\0';

      i = Pa_FindSortAction(&pa_sortTable[c->sort],s);
      free(s);
  
      if (i == -1) {
        Pa_RemoveComposition(numcomp);
        printf(" Action %%%d does not exist.\n",k+1);
        return;
      }
      c->tableA[k++] = i;
    }
    free(sup);
  }

  Compose(c,type);

  /* DEADLOCK STATES */
  /*
  printf("\nDeadlock States:\n");

  trc = c->transitionBDD;

  f1 = Bdd_RelOp(trc,bdd_termFalse,"ExA ExS",TRUE);
  f2 = Bdd_RelOp(trc,bdd_termFalse,"ExA ExR S2R",TRUE);
  f3 = Bdd_ITE(f1,bdd_termFalse,f2);
  Bdd_Fortify(f3.p);

  Bdd_WriteFunction(f3);
  Pa_DecodeCompStates(c,f3,TRUE);
  */

  /*
  ura2 = clock();
  fprintf(stdout,"\nCPU time for composition: %.2f s\n",
          ((float)(ura2-ura1))/((float) CLOCKS_PER_SEC));
  
  Bdd_SystemStat();
  */

  return;
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
Compose(Pa_Composition *c, int type)
{
  int numP;
  int i,j,k;
  Pa_Process *p;
  Bdd_Edge actionsBDD;
  Bdd_Edge tau,stab;
  Bdd_Edge s_new, s_acc;
  Bdd_Edge sout, stau;        /* for normalization */
  Bdd_Edge tr1, tr2;          /* supplemental functions */
  Bdd_Edge sup, sup1, sup2;   /* supplemental functions */
  Bdd_Edge sup2a, sup2b;      /* supplemental functions */
  Bdd_Edge s1, s2, s3;        /* supplemental functions */
  Bdd_Edge new1, new2;        /* supplemental functions */
  Bdd_String pom,s;
  int n,nt;
  Bdd_String name;
  int loop = 0;
  clock_t ura1, ura2;

  ura1 = clock();

  /* NOTE: PROCESSES HAVE TO BE ENCODED WITH DIFFERENT STATE-VARIABLES! */

  /* INITIALIZATION OF SUPPLEMENTAL STRUCTURES */

  numP = c->numProcesses;

  c->di = (Bdd_Edge *) calloc (numP,sizeof(Bdd_Edge));
  c->dij = (Bdd_Edge *) calloc (numP*numP,sizeof(Bdd_Edge));

  c->alphabet = (Bdd_Edge *) calloc (numP,sizeof(Bdd_Edge));
  for(k=0; k<numP; k++) c->alphabet[k] = bdd_termFalse;

  tau = pa_sortTable[c->sort].table[0].p;

  /* COLLECT INFORMATION ABOUT ALPHABET OF PROCESSES */
  for(k=0; k<numP; k++) {
    sup = pa_processTable[c->tableP[k]].d;
    sup = Bdd_E(sup,pa_sortTable[c->sort].a0.p->v);
    c->alphabet[k] = Bdd_RelOp(sup,bdd_termTrue,"#AND Ex xR xS",TRUE);
    Bdd_Fortify(c->alphabet[k].p);
  }

  /* CHECK FOR CONNECTIONS BETWEEN PROCESSES */

  for(i=0; i<numP; i++) c->di[i] = bdd_termFalse;

  for(i=0; i<numP; i++) {
    for(j=0; j<numP; j++) {
      sup = Bdd_ITE(c->alphabet[i],c->alphabet[j],bdd_termFalse);
      if (Bdd_Cmp(sup,bdd_termFalse)) {
        c->dij[i * numP + j] = bdd_termNull;
      } else {
        c->dij[i * numP + j] = bdd_termFalse;
      }
    }
  }

  /* WITH REGARD TO THE TYPE OF COMPOSITION CALCULATE:                 */
  /* 1 OR 4 OR 7 - actionsBDD CONTAINS NON-RESTRICTED ACTIONS AND TAU   */
  /* 2 OR 5 OR 8 - actionsBDD CONTAINS RESTRICTED ACTIONS */

  if ((type == 0) || (type == 3) || (type == 6)) {  /* without restrictions */

    actionsBDD = VersisComposeActions(c);
    Bdd_Fortify(actionsBDD.p);

  }

  if ((type == 1) || (type == 4) || (type == 7)) {

    actionsBDD = VersisComposeActions(c);
    actionsBDD = Bdd_ITE(actionsBDD,bdd_termTrue,tau);  /* add action 'TAU' */
    Bdd_Fortify(actionsBDD.p);

  }

  if ((type == 2) || (type == 5) || (type == 8)) {

    actionsBDD = VersisComposeActions(c);
    Bdd_Fortify(actionsBDD.p);

  }

  Bdd_IncCounter();

  /* CALCULATE INITIAL STATE OF THE COMPOSITION */

  s_new = bdd_termTrue;
  for(i=0; i<numP; i++) {
    p = &pa_processTable[c->tableP[i]];
    s_new = Bdd_ITE(s_new,p->tableS[p->initial].bddR,bdd_termFalse);
  }
  Bdd_Fortify(s_new.p);
  c->initialBDD = s_new;

  /* s_new (RRR) CONTAIN NEW STATES FOUND IN PREVIOUSLY STEP OF CALCULATION */
  /* s_acc (RRR) CONTAIN ALL REACHED STATES */

  s_acc = s_new;

  /* LOOP UNTIL NO NEW STATE IS REACHED */

  do {

    /*
    ura2 = clock();
    fprintf(stdout,"\nComposition loop: %d, CPU: %.2f s\n",++loop,
            ((float)(ura2-ura1))/((float) CLOCKS_PER_SEC));
    Bdd_SystemStat();
    */

    /*
    printf("\nNew states to be processed:\n");
    Pa_DecodeCompStates(c,s_new,TRUE);
    */

    /* PART ONE */
    /* CALCULATE NEW STATES, WHICH ARE REACHED WITHOUT COMMUNICATION */
    /* BETWEEN PROCESSES AND STORE THEM IN new1 (RRR)                */

    new1 = bdd_termFalse;
    for(i=0; i<numP; i++) {

      /* COMPUTE ALL TRANSITIONS IN PROCESS I WHICH ARE ALLOWED   */
      /* WITHOUT COMMUNICATION AND STORE THEM IN s1 (RAS)         */

      if ((type == 1) || (type == 4) || (type == 7)) {
        s1 = Bdd_ITE(actionsBDD,pa_processTable[c->tableP[i]].d,bdd_termFalse);
      }

      if ((type == 0) || (type == 3) || (type == 6) ||
          (type == 2) || (type == 5) || (type == 8)) {
        s1 = Bdd_ITE(actionsBDD,bdd_termFalse,pa_processTable[c->tableP[i]].d);
      }

      /* UPDATE s1 (RRRAS)                                          */
      /* IT NOW CONTAINS ONLY TRANSITIONS STARTING FROM s_new (RRR) */

      s1 = Bdd_ITE(s1,s_new,bdd_termFalse);

      if (type < 3) {

        /* IF STANDARD CCS COMPOSITION IS COMPUTED THEN:           */
        /* 1. ACCUMULATE TRANSITIONS s1 (RRRAS) IN c->di (RRRAS)   */

        c->di[i] = Bdd_ITE(c->di[i],bdd_termTrue,s1);

        /* 2. CALCULATE STATES REACHED WITH TRANSITIONS FROM       */
        /*    s1 (RRRAS) AND STORE THEM IN sup (RRR)               */

        sup = Bdd_NewState1(s1,pa_processTable[c->tableP[i]].name);

        /* 3. ADD STATES FROM sup (RRR) TO new1 (RRR)              */
        new1 = Bdd_ITE(new1,bdd_termTrue,sup);

      } else {

        /* IF MULTI-WAY COMPOSITION WITH ISOCHRONIC FORKS IS COMPUTED THEN: */
        /* 1. CHECK IN OTHER PROCESSES IF THEY CAN PERFORM EQUAL ACTIONS.   */
        /*    STORE COMPLETE TRANSITIONS IN s1 (RRRASSS)                    */

        for(j=0; j<numP; j++) {
          if (j != i) {
            sup1 = Bdd_ITE(s1,pa_processTable[c->tableP[j]].d,bdd_termFalse);
            sup2 = Bdd_ExS(sup1,bdd_termTrue);
            sup2 = Bdd_ITE(sup2,bdd_termFalse,s1);

            /* sup2 CONTAINS TRANSITIONS, WITH WHICH J CANNOT SYNCHRONISE */
            /* sup2a ARE THOSE IN ALPHABET OF J, sup2b ARE OTHERS         */
            sup2a = Bdd_ITE(sup2,c->alphabet[j],bdd_termFalse); 
            sup2b = Bdd_ITE(sup2a,bdd_termFalse,sup2); 
 
            /* TRANSITIONS FROM sup2a ARE WRONG - COMPUTATION INTERFERENCE */
            /* TRANSITIONS FROM sup2b ARE ALLOWED, PROCESS J SIMPLY WAIT   */

            sup1 = Bdd_ITE(sup2a,bdd_termFalse,sup1);
            sup2 = Bdd_ITE(sup2b,pa_processTable[c->tableP[j]].stab,
                           bdd_termFalse);
            s1 = Bdd_ITE(sup1,bdd_termTrue,sup2);
	}}

        /* 2. ACCUMULATE TRANSITIONS s1 (RRRASSS) IN c->di (RRRASSS)  */

        c->di[i] = Bdd_ITE(c->di[i],bdd_termTrue,s1);

      } /* else if (type < 3) */

    } /* for (i) */

    /* PART TWO */
    /* CALCULATE NEW STATES, WHICH ARE REACHED WITH COMMUNICATION */
    /* BETWEEN PROCESSES AND STORE THEM IN new2 (RRR)             */

    new2 = bdd_termFalse;
    for(i=0; i<numP; i++) {

      /* FOR ALL J FIND COMMUNICATION BETWEEN PROCESSES I AND J */

      for(j=0; j<numP; j++)
      if (i != j)
      if (!Bdd_Cmp(c->dij[i * numP + j],bdd_termNull)) {

        /* COMPUTE TRANSITIONS WITHOUT a0  */
	/* IN PROCESS I WITH OUTPUT ACTIONS s1 (RAS)  */
	/* IN PROCESS J WITH INPUT ACTIONS s2 (RAS)   */
        /* ONLY RESTRICTED ACTIONS ARE COMPOSED TOGETHER! */

        s1 = Bdd_Restrict(pa_processTable[c->tableP[i]].d,
                        pa_sortTable[c->sort].a0.p->v, TRUE);

        s2 = Bdd_Restrict(pa_processTable[c->tableP[j]].d,
                        pa_sortTable[c->sort].a0.p->v, FALSE);

        if ((type == 1) || (type == 4) || (type == 7)) { /* not-restricted actions */
          s1 = Bdd_ITE(actionsBDD,bdd_termFalse,s1);
          s2 = Bdd_ITE(actionsBDD,bdd_termFalse,s2);
        }

        if ((type == 2) || (type == 5) || (type == 8)) { /* restricted actions */
          s1 = Bdd_ITE(actionsBDD,s1,bdd_termFalse);
          s2 = Bdd_ITE(actionsBDD,s2,bdd_termFalse);
        }

        if (type < 3) {

          /* IF STANDARD CCS COMPOSITION IS COMPUTED THEN:                  */
          /* 1. COMPOSE s1 (RAS) AND s2 (RAS) TOGETHER AND CREATE s3 (RRSS) */
          /* 2. UPDATE s3 (RRRSS) TO CONTAIN ONLY TRANSITIONS FROM s_new    */

          s3 = Bdd_ExA(s1,s2);
          s3 = Bdd_ITE(s_new,s3,bdd_termFalse);

          /* 3. ACCUMULATE TRANSITIONS s3 (RRRSS) IN c->dij (RRRSS)  */

          c->dij[i * numP + j] = Bdd_ITE(c->dij[i * numP + j],
                                         bdd_termTrue, s3);

          /* 4. CALCULATE STATES REACHED WITH TRANSITIONS FROM   */
          /*    s3 (RRRSS) AND STORE THEM IN sup (RRR)           */

          sup = Bdd_NewState2(s3,
                            pa_processTable[c->tableP[i]].name,
                            pa_processTable[c->tableP[j]].name);

          /* 5. ADD STATES FROM sup (RRR) TO new2 (RRR) */
          new2 = Bdd_ITE(new2,bdd_termTrue,sup);

        } else {

          /* IF MULTI-WAY COMPOSITION WITH ISOCHRONIC FORKS IS COMPUTED THEN:*/
          /* 1. COMPOSE s1 (RAS) AND s2 (RAS) TOGETHER AND CREATE s3 (RRASS) */

          s3 = Bdd_ITE(s1,s2,bdd_termFalse);

          /* 2. CHECK IN OTHER PROCESSES IF THEY CAN PERFORM THE SAME ACTION.*/
          /*    STORE COMPLETE TRANSITIONS IN s3 (RRRASSS)                   */

          for(k=0; k<numP; k++) {
            if ((k != i) && (k != j))
            if ((!Bdd_Cmp(c->dij[i * numP + k],bdd_termNull))
                && (!Bdd_Cmp(c->dij[j * numP + k],bdd_termNull))) {
  
              sup1 = Bdd_ITE(s3,pa_processTable[c->tableP[k]].d,bdd_termFalse);
              sup2 = Bdd_ExS(sup1,bdd_termTrue);
              sup2 = Bdd_E(sup2,pa_sortTable[c->sort].a0.p->v);

              s = (Bdd_String) malloc(255);
              sprintf(s,"Ex xR<%s>",pa_processTable[c->tableP[k]].name);
              sup2 = Bdd_RelOpComplex(sup2,bdd_termTrue,s,TRUE);
              free(s);

              sup2 = Bdd_ITE(sup2,bdd_termFalse,s3);

              /* sup3 = sup1 + sup2 */
              /* sup1 CONTAINS TRANSITIONS, WITH WHICH K CAN SYNCHRONISE     */
              /* sup2 CONTAINS TRANSITIONS, WITH WHICH K CANNOT SYNCHRONISE  */
              sup2a = Bdd_ITE(sup2,c->alphabet[k],bdd_termFalse);
              sup2b = Bdd_ITE(sup2a,bdd_termFalse,sup2);
 
              /* sup2 = sup2a + sup2b */
              /* sup2a ARE THOSE IN ALPHABET OF K, sup2b ARE OTHERS          */
              /* TRANSITIONS FROM sup2a ARE WRONG - COMPUTATION INTERFERENCE */
              /* TRANSITIONS FROM sup2b ARE ALLOWED, PROCESS K SIMPLY WAIT   */

              sup2b = Bdd_ITE(sup2b,pa_processTable[c->tableP[k]].stab,
                              bdd_termFalse);
              s3 = Bdd_ITE(sup1,bdd_termTrue,sup2b);
	    } else {
              s3 = Bdd_ITE(s3,pa_processTable[c->tableP[k]].stab,bdd_termFalse);
	    }
          }

          /* 3. REMOVE ACTION VARIABLES FROM s3                      */
          /* 4. s3 (RRRSSS) TO CONTAIN ONLY TRANSITIONS FROM s_new   */

          s3 = Bdd_ExA(s3,bdd_termTrue);
          s3 = Bdd_ITE(s_new,s3,bdd_termFalse);

          /* 5. ACCUMULATE TRANSITIONS s3 (RRRSS) IN c->dij (RRRSS)  */

          c->dij[i * numP + j] = Bdd_ITE(c->dij[i * numP + j], bdd_termTrue, s3);

        }  /* else if (type < 3) */

      } /* for (j) */

    } /* for (i) */

    /* MULTI-WAY COMPOSITION WITH ISOCHRONIC FORKS */

    if (type > 2) {

      if ((type == 6) || (type == 7) || (type == 8)) {

        /* PREPARING FUNDAMENTAL-MODE NORMALIZATION */

        /* FIND STATES WITH OUTPUT TRANSITION */
        sout = bdd_termFalse;
        for(i=0; i<numP; i++) {

          sup = Bdd_RelOp(c->di[i],
                          pa_sortTable[c->sort].a0,"#AND Ex xA xS",TRUE);
          sout = Bdd_ITE(sout,bdd_termTrue,sup);

        } /* for (i) */

        /*
        printf("\nReached states with output transition:\n");
        Pa_DecodeCompStates(c,sout,TRUE);
        */

        /* FIND STATES WITH TAU TRANSITION */
        stau = bdd_termFalse;
        for(i=0; i<numP; i++) {
          for(j=0; j<numP; j++)
          if (i != j)
          if (!Bdd_Cmp(c->dij[i * numP + j],bdd_termNull)) {

            sup = Bdd_RelOp(c->dij[i * numP + j],bdd_termTrue,
                            "#AND Ex xS",TRUE);
            stau = Bdd_ITE(stau,bdd_termTrue,sup);

        }} /* for (i) for (j) */

        /*
        printf("\nReached states with tau transition:\n");
        Pa_DecodeCompStates(c,stau,TRUE);
        */

      } /* END OF PREPARING NORMALIZATION */

      /* CALCULATION OF NEW STATES */

      new1 = bdd_termFalse;
      for(i=0; i<numP; i++) {

        /* CALCULATE STATES REACHED WITH TRANSITIONS c->di[i] (RRRASSS) */
        /* AND STORE THEM IN sup (RRR)                                  */

        sup1 = Bdd_ITE(pa_sortTable[c->sort].a0,c->di[i],bdd_termFalse); /*!*/
        sup2 = Bdd_ITE(pa_sortTable[c->sort].a0,bdd_termFalse,c->di[i]); /*?*/

        /* FUNDAMENTAL-MODE NORMALIZATION */
        if ((type == 6) || (type == 7) || (type == 8)) {
          sup2 = Bdd_ITE(sout,bdd_termFalse,sup2);
          sup2 = Bdd_ITE(stau,bdd_termFalse,sup2);
	}

        c->di[i] = Bdd_ITE(sup1,bdd_termTrue,sup2);

        sup = Bdd_NewState(c->di[i]);

        /* ADD STATES FROM sup (RRR) TO new1 (RRR) */
        new1 = Bdd_ITE(new1,bdd_termTrue,sup);

      } /* for (i) */

      new2 = bdd_termFalse;
      for(i=0; i<numP; i++) {
        for(j=0; j<numP; j++)
	if (i != j)
        if (!Bdd_Cmp(c->dij[i * numP + j],bdd_termNull)) {

          /* CALCULATE STATES REACHED WITH TRANSITIONS FROM   */
          /* c->dij (RRRSS) AND STORE THEM IN sup (RRR)       */

          /* NORMALIZATION */
          if ((type == 6) || (type == 7) || (type == 8)) {
            c->dij[i * numP + j] =
                   Bdd_ITE(sout,bdd_termFalse,c->dij[i * numP + j]);
	  }

          sup = Bdd_NewState(c->dij[i * numP + j]);

          /* ADD STATES FROM sup (RRR) TO new2 (RRR) */
          new2 = Bdd_ITE(new2,bdd_termTrue,sup);

        }
      }
    }

    /*
    printf("\nNew states found and stored in new1:\n");
    Pa_DecodeCompStates(c,new1,TRUE);
    printf("\nNew states found and stored in new2:\n");
    Pa_DecodeCompStates(c,new2,TRUE);
    */

    /* UPDATE s_new (RRR) */
    s_new = Bdd_ITE(new1,bdd_termTrue,new2);
    s_new = Bdd_ITE(s_acc,bdd_termFalse,s_new);

    /* UPDATE s_acc (RRR) */
    s_acc = Bdd_ITE(s_acc,bdd_termTrue,s_new);

    Bdd_IncCounter();
    Bdd_Fresh(s_new.p);
    Bdd_Fresh(s_acc.p);

    for(i=0; i<numP; i++) {
      Bdd_Fresh(c->di[i].p);
      for(j=0; j<numP; j++)
      if (!Bdd_Cmp(c->dij[i * numP + j],bdd_termNull)) {
        Bdd_Fresh(c->dij[i * numP + j].p);
      }
    }

  } while(!(Bdd_Cmp(s_new,bdd_termFalse)));

  Bdd_Fortify(s_acc.p);
  for(i=0; i<numP; i++) {
    Bdd_Fortify(c->di[i].p);
    for(j=0; j<numP; j++)
    if (!Bdd_Cmp(c->dij[i * numP + j],bdd_termNull)) {
      Bdd_Fortify(c->dij[i * numP + j].p);
    }
  }

  if (type < 3) {

    /* IF STANDARD CCS COMPOSITION IS COMPUTED di AND dij */
    /* WITH ADDED stab MUST BE COLLECTED                  */

    /* ADD stab AND COLLECT di */

    tr1 = bdd_termFalse;
    for(i=0; i<numP; i++) {
      sup = bdd_termTrue;
      for(k=0; k<numP; k++) {
        if (k != i)
          sup = Bdd_ITE(sup,pa_processTable[c->tableP[k]].stab,bdd_termFalse);
      }
      sup = Bdd_ITE(sup,c->di[i],bdd_termFalse);
      tr1 = Bdd_ITE(tr1,bdd_termTrue,sup);

      /* THIS Bdd_IncCounter() IS OPTIONAL */
      /**/
      Bdd_IncCounter();
      Bdd_Fresh(tr1.p);
      /**/
    }

    /* ADD stab AND COLLECT dij */

    tr2 = bdd_termFalse;
    for(i=0; i<numP; i++) {
      for(j=0; j<i; j++)
      if (!Bdd_Cmp(c->dij[i * numP + j],bdd_termNull)) {
        sup = bdd_termTrue;
        for(k=0; k<numP; k++) {
          if ((k != i) && (k != j))
            sup = Bdd_ITE(sup,pa_processTable[c->tableP[k]].stab,bdd_termFalse);
        }
        sup1 = Bdd_ITE(sup,c->dij[i * numP + j],bdd_termFalse);
        sup2 = Bdd_ITE(sup,c->dij[j * numP + i],bdd_termFalse);
        tr2 = Bdd_ITE(tr2,bdd_termTrue,sup1);
        tr2 = Bdd_ITE(tr2,bdd_termTrue,sup2);
      }

      /* THIS Bdd_IncCounter() IS OPTIONAL */
      /**/
      Bdd_IncCounter();
      Bdd_Fresh(tr1.p);
      Bdd_Fresh(tr2.p);
      /**/
    }

  } else {

    /* IF MULTI-WAY COMPOSITION WITH ISOCHRONIC FORKS IS COMPUTED */
    /* di AND dij MUST BE SIMPLY COLLECTED                        */ 

    /* COLLECT di */

    tr1 = bdd_termFalse;
    for(i=0; i<numP; i++) {
      tr1 = Bdd_ITE(tr1,bdd_termTrue,c->di[i]);
    }

    /* ADD stab AND COLLECT dij */

    tr2 = bdd_termFalse;
    for(i=0; i<numP; i++) {
      for(j=0; j<i; j++)
      if (!Bdd_Cmp(c->dij[i * numP + j],bdd_termNull)) {
        tr2 = Bdd_ITE(tr2,bdd_termTrue,c->dij[i * numP + j]);
        tr2 = Bdd_ITE(tr2,bdd_termTrue,c->dij[j * numP + i]);
      }
    }

  } /* if (type) */

  tr2 = Bdd_ITE(tau,tr2,bdd_termFalse); /* tr2 = 'tau' * tr2 */
  tr1 = Bdd_ITE(tr1,bdd_termTrue,tr2);

  Bdd_Fortify(tr1.p);
  Bdd_IncCounter();

  c->stateBDD = s_acc;
  c->transitionBDD = tr1;

  /* BUILD stab BDD for composed process */

  stab = bdd_termTrue;
  for(i=0; i<numP; i++) {
    stab = Bdd_ITE(stab,pa_processTable[c->tableP[i]].stab,bdd_termFalse);
  }
  Bdd_Fortify(stab.p);
  Bdd_IncCounter();
  c->stab = stab;

  /* SAVE RESULTS */

  Bdd_SaveFormula("D_COMP_",c->name,c->transitionBDD);
  Bdd_SaveFormula("Initial_COMP_",c->name,c->initialBDD);
  Bdd_SaveFormula("S__COMP",c->name,c->stateBDD);

  if (type > 2)
  {

    /*
    fprintf(stdout,"\nStatistic after Composition\n");
    fprintf(stdout,"About composition:\n");
    Bdd_FunctionStat(c->transitionBDD);
    fprintf(stdout,"About system:\n");
    Bdd_SystemStat();
    */

    /* IF COMPOSITION IS NOT NORMALIZED, THIS CAN BE DONE HERE */
    /*
    Normalization(c);
    */

    /* IF MULTI-WAY COMPOSITION WITH ISOCHRONIC FORKS IS COMPUTED */
    /* THEN TRACE MINIMIZATION OF COMPOSITION IS PERFORMED */
    /* THE RESULT IS A PROCESS, COMPOSITION IS DELETED! */

    name = strdup(c->name);
    n = Pa_FindComposition(name);                   /* find composition   */
    nt = VersisTraceMinimization(0,name,1,n);       /* minimize it        */
    Pa_RemoveComposition(n);                        /* remove composition */
    free(name);

    /*
    fprintf(stdout,"\nStatistic after Minimization\n");
    Bdd_SystemStat();
    */
  }

  /* FORCE GARBAGE COLLECTION - USED FOR TESTING */

  /*
  Bdd_Garbage();
  */

  /*
  ura2 = clock();
  fprintf(stdout,"\nCPU time for composition: %.2f s\n",
          ((float)(ura2-ura1))/((float) CLOCKS_PER_SEC));
  
  */

  /* POMOZNI IZPISI */

  /*
  printf("\n");
  Pa_DecodeCompStates(c,c->stateBDD,FALSE);
  Pa_DecodeCompTR(c,c->transitionBDD,FALSE);
  Pa_DecodeCompTR(c,Bdd_ITE(tau,bdd_termFalse,c->transitionBDD),FALSE);
  printf("\n");
  */

}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
Normalization(Pa_Composition *c)
{
  Bdd_Edge tau;
  Bdd_Edge input,output,internal;
  Bdd_Edge intauout,intau;

  tau = pa_sortTable[c->sort].table[0].p;

  output = Bdd_RelOp(c->transitionBDD,
                      pa_sortTable[c->sort].a0,"#AND Ex xA xS",TRUE);

  /*
  printf("\nStates with an output transition:\n");
  Pa_DecodeCompStates(c,output,TRUE);
  */

  internal = Bdd_RelOp(c->transitionBDD,tau,"#AND Ex xA xS",TRUE);

  /*
  printf("\nStates with tau transition:\n");
  Pa_DecodeCompStates(c,internal,TRUE);
  */

  input = Bdd_ITE(pa_sortTable[c->sort].a0,bdd_termFalse,c->transitionBDD);

  /*
  printf("\nAll input transitions and tau transitions:\n");
  Pa_DecodeCompTR(c,input,TRUE);
  */

  intauout = Bdd_ITE(input,output,bdd_termFalse);

  /*
  printf("\nInput and tau transitions parallel to an output transition:\n");
  Pa_DecodeCompTR(c,intauout,TRUE);
  */

  input = Bdd_ITE(tau,bdd_termFalse,input);
  intau = Bdd_ITE(input,internal,bdd_termFalse);

  /*
  printf("\nInput transitions parallel to tau transition:\n");
  Pa_DecodeCompTR(c,intau,TRUE);
  */

  c->transitionBDD = Bdd_ITE(intauout,bdd_termFalse,c->transitionBDD);
  c->transitionBDD = Bdd_ITE(intau,bdd_termFalse,c->transitionBDD);

  Bdd_Fortify(c->transitionBDD.p);
}


/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
wordNumber(Bdd_String s)
{
  int num;
  int token;
  Bdd_String sup,sup_next;

  if (!s) return 0;

  num = 0;
  sup = strdup(s);

  token = strspn(sup," ");
  while (strspn(sup," ") < strlen(sup)) {
    sup_next = (Bdd_String) strdup(&sup[token]);
    free(sup);
    sup = sup_next;

    num++;
    token = strcspn(sup," ");
    sup_next = (Bdd_String) strdup(&sup[token]);
    free(sup);
    sup = sup_next;

    token = strspn(sup," ");
  }

  free(sup);
  return num;
}
