/**CFile*******************************************************************
  PackageName [versis]
  Synopsis    [Package 'versis' provides basic tools for formal
               verification of concurrent systems]

  FileName    [versisMin.c]
  Revision    [$Revision: 2 $]
  Date        [$Date: 2006-10-05 15:30:29 +0200 (Thu, 05 Oct 2006) $]
  Authors     [Robert Meolic (meolic@uni-mb.si),
               Tatjana Kapus (kapus@uni-mb.si)]
  Description [File versisMin.c provides different minimizations.]
  SeeAlso     [versis.h, versisInt.h]

  Copyright   [see versis.h]
  ************************************************************************/

#include "versisInt.h"

/*-----------------------------------------------------------------------*/
/* Structure declarations                                                */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static Bdd_Boolean Minimization(Bdd_String pref, Bdd_Edge eqv,
                                int par, int n, int type);

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Function Versis_Minimization.]
  Description [type = 0,1,3,4 - strong, weak, trace, trace_strong
               type = 2 - testing (not implemented)
               type = 5 - minimization of WCA]
  SideEffects [trace minimization returns minimal deterministic LTS]
  SeeAlso     []
  ************************************************************************/

int
Versis_Minimization(int type, int par, Bdd_String name)
{
  int n;
  Bdd_String newname;
  int result;

  if (par == 0) {
    n = Pa_FindProcess(name);
    if (n == -1) {
      printf(" Proces %s does not exist.\n",name);
      return -1;
    }
    if ( (!pa_processTable[n].encoded) && (type != 5) ) {
      printf(" Proces %s is not encoded.\n",name);
      return -1;
    }
  }
  if (par == 1) {
    n = Pa_FindComposition(name);
    if (n == -1) {
      printf(" Composition %s does not exist.\n",name);
      return -1;
    }
  }
  
  switch (type) {
    case 0: {
              newname = strdup("STRONG_");
              newname = (Bdd_String)
                            realloc(newname,strlen(newname)+strlen(name)+1);
              strcat(newname,name);

              result = VersisStrongMinimization(newname,par,n);

              free(newname);
              break;
            }

    case 1: {
              newname = strdup("WEAK_");
              newname = (Bdd_String)
                            realloc(newname,strlen(newname)+strlen(name)+1);
              strcat(newname,name);

              result = VersisWeakMinimization(newname,par,n);

              free(newname);
              break;
            }

    case 2: {
              break;
            }

    case 3: {
              newname = strdup("TRACE_");
              newname = (Bdd_String)
                            realloc(newname,strlen(newname)+strlen(name)+1);
              strcat(newname,name);

              result = VersisTraceMinimization(0,newname,par,n);

              free(newname);
              break;
            }

    case 4: {
              newname = strdup("TRACE_STRONG_");
              newname = (Bdd_String)
                            realloc(newname,strlen(newname)+strlen(name)+1);
              strcat(newname,name);

              result = VersisTraceMinimization(1,newname,par,n);

              free(newname);
              break;
            }

    case 5: {
              newname = strdup("MIN_");
              newname = (Bdd_String)
                            realloc(newname,strlen(newname)+strlen(name)+1);
              strcat(newname,name);

              result = VersisWCAMinimization(newname,par,n);

              free(newname);
              break;
            }
  }

  return result;
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
VersisStrongMinimization(Bdd_String name, int par, int n)
{
  Bdd_Edge eqv;

  VersisStrongEquivalence(&eqv,par,n,-1,-1);

  /*
  Pa_DecodeProcProcPair(&pa_processTable[n],&pa_processTable[n],eqv,TRUE);
  */

  Minimization(name,eqv,par,n,0);

  return Pa_FindProcess(name);
}


/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
VersisWeakMinimization(Bdd_String name, int par, int n)
{
  Bdd_Edge eqv;

  VersisWeakEquivalence(&eqv,par,n,-1,-1);

  /*
  Pa_DecodeProcProcPair(&pa_processTable[n],&pa_processTable[n],eqv,TRUE);
  */

  Minimization(name,eqv,par,n,1);

  return Pa_FindProcess(name);
}


/**Function****************************************************************
  Synopsis    []
  Description [type: 0 = weak = tau is an internal action,
                     1 = strong = tau is not a special action]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
VersisTraceMinimization(int type, Bdd_String name, int par, int n)
{
  Bdd_Edge D;
  int s;
  Bdd_Edge tau, stab, init, Ptau, trcl;
  Bdd_String inname, name1, orgname;
  Bdd_Edge pom, eps, eps_rp, eps_qs;
  Bdd_Edge eqv;
  int det;
  int i;

  if (par == 0) {
    D = pa_processTable[n].d;
    inname = pa_processTable[n].name;
    orgname = strdup(inname);
    s = pa_processTable[n].sort;      /* sort */
    tau = pa_sortTable[s].table[0].p; /* tau */
    stab = pa_processTable[n].stab;
    init = pa_processTable[n].tableS[pa_processTable[n].initial].bddR;
  } else {
    D = pa_compositionTable[n].transitionBDD;
    inname = pa_compositionTable[n].name;
    orgname = NULL;
    s = pa_compositionTable[n].sort;  /* sort */
    tau = pa_sortTable[s].table[0].p; /* tau */
    stab = pa_compositionTable[n].stab;
    init = pa_compositionTable[n].initialBDD;
  }

  /* NAME OF DETERMINISTIC PROCESS */
  name1 = (Bdd_String) strdup("DET_");
  name1 = (Bdd_String) realloc(name1,strlen(name1)+strlen(inname)+1);
  strcat(name1,inname);

  /* dolocimo Ptau - kam vse lahko gres iz posameznega stanja */
  /* eps so tau prehodi v D */

  if (type == 0) {
    pom = Bdd_ExA(D,tau);
    trcl = VersisTrCl(pom);
    Bdd_Fortify(trcl.p);
    Bdd_IncCounter();

    pom = Bdd_R2P_S2Q(D);
    eps = Bdd_ITE(trcl,bdd_termTrue,stab);
    Bdd_IncCounter();
    Bdd_Fresh(pom.p);
    Bdd_Fresh(eps.p);

    eps_rp = Bdd_S2P(eps);
    pom = Bdd_ExP(eps_rp,pom);          /* z leve mnozimo s tau prehodi */
    Bdd_IncCounter();
    Bdd_Fresh(eps.p);
    Bdd_Fresh(pom.p);

    eps_qs = Bdd_R2Q(eps);
    pom = Bdd_ExQ(pom,eps_qs);         /* z desne mnozimo s tau prehodi */
    Bdd_IncCounter();
    Bdd_Fresh(pom.p);

    Ptau = Bdd_ITE(tau,bdd_termFalse,pom);       /* brisemo tau prehode */
    Bdd_Fortify(Ptau.p);
    Bdd_IncCounter();
  }

  if (type == 1) {
    trcl = bdd_termFalse;
    Ptau = D;
  }

  VersisBuildDet(orgname,name1,init,Ptau,trcl,s);
  det = Pa_FindProcess(name1);

  if (name1) free(name1);
  if (orgname) free(orgname);

  /* CHECK THE NUMBER OF STATES IN det - BEFORE MINIMIZATION */
  /*
  printf(" (BEFORE)");
  i = Pa_DecodeProcessStates(&pa_processTable[det],pa_processTable[det].stateBDD,FALSE);
  printf(" (%s: %d st, ",pa_processTable[det].name,i);
  i = Pa_DecodeProcessTR(&pa_processTable[det],pa_processTable[det].d,FALSE);
  printf("%d tr)...",i);
  */

  /* THIS METHOD FINDS A PROCESS WHICH IS NOT MINIMAL      */
  /* IT IS BISIMULATION EQUIVALENT TO THE MINIMAL PROCESS  */
  /*
  name1 = strdup(pa_processTable[det].name);
  Pa_CopyProcess(name1,name,NULL,"","");
  Pa_DeleteProcess(Pa_FindProcess(name1));
  free(name1);
  */

  /* THIS METHOD FINDS MINIMAL DETERMINISTIC PROCESS  */
  /* IT IS MUCH MORE COMPLEX THAN THE PREVIOUS METHOD */
  /**/
  VersisStrongEquivalence(&eqv,0,det,-1,-1);
  Minimization(name,eqv,0,det,3);
  /**/

  Pa_DeleteProcess(det);
  det = Pa_FindProcess(name);

  /* CHECK THE NUMBER OF STATES IN det - AFTER MINIMIZATION */
  /*
  printf(" (AFTER)");
  i = Pa_DecodeProcessStates(&pa_processTable[det],pa_processTable[det].stateBDD,FALSE);
  printf(" (%s: %d st, ",pa_processTable[det].name,i);
  i = Pa_DecodeProcessTR(&pa_processTable[det],pa_processTable[det].d,FALSE);
  printf("%d tr)...",i);
  */

  return det;
}


/**Function****************************************************************
  Synopsis    []
  Description [Remove duplicated paths from WCA]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
VersisWCAMinimization(Bdd_String name, int par, int n)
{
  Bdd_Edge D,init;
  int s;
  int final;
  Bdd_Edge eqv;
  int i,m;

  /* MINIMIZATION IS WORKING ON A COPY OF WCA */

  n = Pa_CopyProcess(pa_processTable[n].name, "_TMP", NULL, "", "");

  /* MARK FINAL STATES WITH ADDITIONAL TRANSITIONS */
  final = Pa_FOAStateProcess(&pa_processTable[n],"FINAL");
  for(i=0; i<pa_processTable[n].numStates; i++) {
    if (pa_processTable[n].tableS[i].final) {
      Pa_FOATransition(&pa_processTable[n],
                   i,Pa_FOASortAction(&pa_sortTable[pa_processTable[n].sort],"FINAL"),1,final);
    }
  }

  Pa_EncodeProcess("_TMP");
  m = VersisTraceMinimization(0,name,0,n);

  /* REMOVE ADDITIONAL TRANSITIONS */
  Versis_ChangeForbid(pa_processTable[m].name,"FINAL");

  /* REMOVE TMP COPY OF THE WCA */
  Pa_DeleteProcess(n);

  return Pa_FindProcess(name);
}

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Minimization]
  Description [type: 0 - strong, 1 - weak , 3 - trace minimization]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Boolean
Minimization(Bdd_String name, Bdd_Edge eqv, int par, int n, int type)
{
  Bdd_Edge D,init,sall;
  Bdd_String x;
  int s,min,t;
  Bdd_Edge sup,sup1,sup2,new,set;
  Bdd_Edge a;
  int i,j,k;
  Bdd_String pom;
  Bdd_Edge final;

  if (par == 0) {
    s = pa_processTable[n].sort; /* sort */
    D = pa_processTable[n].d;
    init = pa_processTable[n].tableS[pa_processTable[n].initial].bddR;

    final = bdd_termFalse;
    for(i=0; i<pa_processTable[n].numStates; i++) {
      if (pa_processTable[n].tableS[i].final) {
        final = Bdd_ITE(final,bdd_termTrue,pa_processTable[n].tableS[i].bddR);
      }
    }
    Bdd_Fortify(final.p);

  } else {
    s = pa_compositionTable[n].sort; /* sort */
    D = pa_compositionTable[n].transitionBDD;
    init = pa_compositionTable[n].initialBDD;
    final = bdd_termFalse;
  }

  /* we will need the set of all states */
  sup = Bdd_RelOp(D,bdd_termFalse,"ExA ExS",TRUE);
  sup1 = Bdd_RelOp(D,bdd_termFalse,"ExA ExR S2R",TRUE);
  sall = Bdd_ITE(sup,bdd_termTrue,sup1);
  Bdd_Fortify(sall.p);

  /* tvorimo ogrodje novega procesa */
  min = Pa_AddNewProcess(name);
  pa_processTable[min].sort = s;
  pa_processTable[min].numStates= 0;
  pa_processTable[min].initial = 0;

  /* tvorimo zacetno stanje */

  x = strdup("s");
  t = Pa_FOANextStateProcess(&pa_processTable[min],x);
  free(x);

  /* bddR in each state temporaly represents the set of original states */
  sup = Bdd_R2P(init);
  sup = Bdd_ExP(eqv,sup);
  Bdd_Fortify(sup.p);

  pa_processTable[min].tableS[t].bddR = sup;

  set = Bdd_ITE(sup,bdd_termFalse,sall);

  /*
  printf("\nIzracun stanj...\n");
  */

  /* v vsakem koraku poiscemo eno novo stanje minimalnega procesa */
  while (!Bdd_Cmp(set,bdd_termFalse)) {

    /*
    pom = (Bdd_String) malloc(255);
    sprintf(pom," Stanje: %d\n",t);
    printf(pom);
    free(pom);
    */

    if (par == 0) {
      new = Pa_ExtractProcState(&pa_processTable[n],set);
    } else {
      new = Pa_ExtractCompState(&pa_compositionTable[n],set);
    }

    x = strdup("s");
    t = Pa_FOANextStateProcess(&pa_processTable[min],x);
    free(x);

    /* bddR in each state temporaly represents the set of original states */
    sup = Bdd_R2P(new);
    sup = Bdd_RelOp(eqv,sup,"#AND Ex xP",TRUE);
    Bdd_Fortify(sup.p);
    pa_processTable[min].tableS[t].bddR = sup;

    set = Bdd_ITE(sup,bdd_termFalse,set);
  }

  /* dolocimo koncna stanja v minimalnem procesu */
  /* stanje je koncno, ce je VSAJ ENO izmed stanj v ekvivalencnem razredu koncno! */

  for (i=0; i<=t; i++) {
    sup = Bdd_ITE(pa_processTable[min].tableS[i].bddR,final,bdd_termFalse);
    if (!Bdd_Cmp(sup,bdd_termFalse)) {
      pa_processTable[min].tableS[i].final = TRUE;
    }
  }

  /* dolocimo prehode v minimalnem procesu */

  /*
  printf("Izracun prehodov...\n");
  */

  for (i=0; i<=t; i++) {
    sup = pa_processTable[min].tableS[i].bddR;

    for (j=0; j<pa_sortTable[s].numActions; j++) {
      a = pa_sortTable[s].table[j].p;

      /*
      pom = (Bdd_String) malloc(255);
      sprintf(pom," Stanje: %d, Akcija: %d\n",i,j);
      printf(pom);
      free(pom);
      */

      /* output action */

      sup1 = Bdd_ITE(pa_sortTable[s].a0,a,bdd_termFalse);
      sup1 = Bdd_ITE(sup,sup1,bdd_termFalse);
      sup1 = Bdd_ITE(D,sup1,bdd_termFalse);
      sup1 = Bdd_NewState(sup1);

      for (k=0; k<=t; k++) {
        sup2 = Bdd_ITE(pa_processTable[min].tableS[k].bddR,sup1,bdd_termFalse);
        if (!Bdd_Cmp(sup2,bdd_termFalse)) {
          Pa_FOATransition(&pa_processTable[min], i, j, 1, k);
        }
      }

      /* input action & action TAU */

      sup1 = Bdd_ITE(pa_sortTable[s].a0,bdd_termFalse,a);
      sup1 = Bdd_ITE(sup,sup1,bdd_termFalse);
      sup1 = Bdd_ITE(D,sup1,bdd_termFalse);
      sup1 = Bdd_NewState(sup1);

      for (k=0; k<=t; k++) {
        if ((type!=1) || (j!=0) || (i!=k)) {
          sup2 = Bdd_ITE(pa_processTable[min].tableS[k].bddR,sup1,bdd_termFalse);
          if (!Bdd_Cmp(sup2,bdd_termFalse)) {
            Pa_FOATransition(&pa_processTable[min], i, j, 0, k);
          }
	}
      }

    }
  }

  /* encode process */
  Pa_EncodeProcess(name);
  Bdd_IncCounter();

  return TRUE;
}
